"""State module for managing Amazon Application Autoscaling scalable target."""
import copy
import re
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    service_namespace: str,
    scaling_resource_id: str,
    scalable_dimension: str,
    resource_id: str = None,
    min_capacity: int = None,
    max_capacity: int = None,
    role_arn: str = None,
    suspended_state: make_dataclass(
        "SuspendedState",
        [
            ("DynamicScalingInSuspended", bool, field(default=None)),
            ("DynamicScalingOutSuspended", bool, field(default=None)),
            ("ScheduledScalingSuspended", bool, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Registers or updates a scalable target.

    A scalable target is a resource that Application Auto Scaling can scale
    out and scale in. Scalable targets are uniquely identified by the combination of resource ID, scalable
    dimension, and namespace.

    When you register a new scalable target, you must specify values for minimum and
    maximum capacity. Current capacity will be adjusted within the specified range when scaling starts. Application
    Auto Scaling scaling policies will not scale capacity to values that are outside of this range.

    After you register a scalable target, you do not need to register it again to use other Application Auto Scaling
    operations. To see which resources have been registered, use DescribeScalableTargets. You can also view the
    scaling policies for a service namespace by using DescribeScalableTargets. If you no longer need a scalable
    target, you can deregister it by using DeregisterScalableTarget.

    To update a scalable target, specify the
    parameters that you want to change. Include the parameters that identify the scalable target: resource ID,
    scalable dimension, and namespace. Any parameters that you don't specify are not changed by this update request.

    .. Note::
        If you call the RegisterScalableTarget API to update an existing scalable target, Application Auto Scaling
        retrieves the current capacity of the resource. If it is below the minimum capacity or above the maximum
        capacity, Application Auto Scaling adjusts the capacity of the scalable target to place it within these bounds,
        even if you don't include the MinCapacity or MaxCapacity request parameters.


    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        service_namespace(str):
            The namespace of the Amazon Web Services service that provides the resource. For a resource
            provided by your own application or service, use custom-resource instead.

        scaling_resource_id(str):
            The identifier of the resource that is associated with the scalable target. This string consists
            of the resource type and unique identifier.

            * ECS service - The resource type is service and the
              unique identifier is the cluster name and service name. Example: service/default/sample-webapp.
            * Spot Fleet - The resource type is spot-fleet-request and the unique identifier is the Spot Fleet
              request ID. Example: spot-fleet-request/sfr-73fbd2ce-aa30-494c-8788-1cee4EXAMPLE.
            * EMR cluster - The resource type is instancegroup and the unique identifier is the cluster ID and instance
              group ID. Example: instancegroup/j-2EEZNYKUA1NTV/ig-1791Y4E1L8YI0.
            * AppStream 2.0 fleet - The resource type is fleet and the unique identifier is the fleet name.
              Example: fleet/sample-fleet.
            * DynamoDB table - The resource type is table and the unique identifier is the table name.
              Example: table/my-table.
            * DynamoDB global secondary index - The resource type is index and the
              unique identifier is the index name. Example: table/my-table/index/my-table-index.
            * Aurora DB cluster - The resource type is cluster and the unique identifier is the cluster name. Example:
              cluster:my-db-cluster.
            * SageMaker endpoint variant - The resource type is variant and the
              unique identifier is the resource ID. Example: endpoint/my-end-point/variant/KMeansClustering.
            * Custom resources are not supported with a resource type.
              This parameter must specify the OutputValue from the CloudFormation template stack used to access
              the resources. The unique identifier is defined by the service provider.
              More information is available in our GitHub repository.
            * Amazon Comprehend document classification endpoint - The resource type and unique
              identifier are specified using the endpoint ARN.
              Example: arn:aws:comprehend:us-west-2:123456789012:document-classifier-endpoint/EXAMPLE.
            * Amazon Comprehend entity recognizer endpoint - The resource type and unique identifier are specified
              using the endpoint ARN.
              Example: arn:aws:comprehend:us-west-2:123456789012:entity-recognizer-endpoint/EXAMPLE.
            * Lambda provisioned concurrency - The resource type is function and the unique identifier is the
              function name with a function version or alias name suffix that is not $LATEST. Example:
              function:my-function:prod or function:my-function:1.
            * Amazon Keyspaces table - The resource type is table and the unique identifier is the table name.
              Example: keyspace/mykeyspace/table/mytable.
            * Amazon MSK cluster - The resource type and unique identifier are specified using the cluster ARN.
              Example: arn:aws:kafka:us-east-1:123456789012:cluster/demo-cluster-1
              /6357e0b2-0e6a-4b86-a0b4-70df934c2e31-5.
            * Amazon ElastiCache replication group - The resource type is replication-group and the unique identifier
              is the replication group name. Example: replication-group/mycluster.
            * Neptune cluster - The resource type is cluster and the unique identifier is the cluster name. Example:
              cluster:mycluster.

        scalable_dimension(str):
            The scalable dimension associated with the scalable target. This string consists of the service
            namespace, resource type, and scaling property.

            * ecs:service:DesiredCount - The desired task count of an ECS service.
            * elasticmapreduce:instancegroup:InstanceCount - The instance count of an EMR Instance Group.
            * ec2:spot-fleet-request:TargetCapacity - The target capacity of a Spot Fleet.
            * appstream:fleet:DesiredCapacity - The desired capacity of an AppStream 2.0 fleet.
            * dynamodb:table:ReadCapacityUnits - The provisioned read capacity for a DynamoDB table.
            * dynamodb:table:WriteCapacityUnits - The provisioned write capacity for a DynamoDB table.
            * dynamodb:index:ReadCapacityUnits - The provisioned read capacity for a DynamoDB global secondary index.
            * dynamodb:index:WriteCapacityUnits - The provisioned write capacity for a DynamoDB global secondary index.
            * rds:cluster:ReadReplicaCount - The count of Aurora Replicas in an Aurora DB cluster.
              Available for Aurora MySQL-compatible edition and Aurora PostgreSQL- compatible edition.
            * sagemaker:variant:DesiredInstanceCount -
              The number of EC2 instances for an SageMaker model endpoint variant.
            * custom-resource:ResourceType:Property -
              The scalable dimension for a custom resource provided by your own application or service.
            * comprehend:document-classifier-endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend document classification endpoint.
            * comprehend:entity-recognizer-endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend entity recognizer endpoint.
            * lambda:function:ProvisionedConcurrency - The provisioned concurrency for a Lambda function.
            * cassandra:table:ReadCapacityUnits - The provisioned read capacity for an Amazon Keyspaces table.
            * cassandra:table:WriteCapacityUnits - The provisioned write capacity for an Amazon Keyspaces table.
            * kafka:broker-storage:VolumeSize -
              The provisioned volume size (in GiB) for brokers in an Amazon MSK cluster.
            * elasticache:replication-group:NodeGroups -
              The number of node groups for an Amazon ElastiCache replication group.
            * elasticache:replication-group:Replicas -
              The number of replicas per node group for an Amazon ElastiCache replication group.
            * neptune:cluster:ReadReplicaCount - The count of read replicas in an Amazon Neptune DB cluster.

        min_capacity(int, Optional):
            The minimum value that you plan to scale in to. When a scaling policy is in effect, Application
            Auto Scaling can scale in (contract) as needed to the minimum capacity limit in response to
            changing demand. This property is required when registering a new scalable target.

            For certain resources, the minimum value allowed is 0. This includes Lambda provisioned concurrency, Spot
            Fleet, ECS services, Aurora DB clusters, EMR clusters, and custom resources. For all other
            resources, the minimum value allowed is 1. Defaults to None.

        max_capacity(int, Optional):
            The maximum value that you plan to scale out to. When a scaling policy is in effect, Application
            Auto Scaling can scale out (expand) as needed to the maximum capacity limit in response to
            changing demand. This property is required when registering a new scalable target.

            Although you can specify a large maximum capacity, note that service quotas may impose lower limits. Each
            service has its own default quotas for the maximum capacity of the resource. If you want to
            specify a higher limit, you can request an increase. For more information, consult the
            documentation for that service. For information about the default quotas for each service, see
            Service Endpoints and Quotas in the Amazon Web Services General Reference. Defaults to None.

        role_arn(str, Optional):
            This parameter is required for services that do not support service-linked roles (such as Amazon
            EMR), and it must specify the ARN of an IAM role that allows Application Auto Scaling to modify
            the scalable target on your behalf.

            If the service supports service-linked roles, Application
            Auto Scaling uses a service-linked role, which it creates if it does not yet exist. For more
            information, see Application Auto Scaling IAM roles. Defaults to None.

        suspended_state(dict[str, Any], Optional):
            An embedded object that contains attributes and attribute values that are used to suspend and
            resume automatic scaling. Setting the value of an attribute to true suspends the specified
            scaling activities. Setting it to false (default) resumes the specified scaling activities.

            Suspension Outcomes

            * For DynamicScalingInSuspended, while a suspension is in effect, all
              scale-in activities that are triggered by a scaling policy are suspended.
            * For DynamicScalingOutSuspended, while a suspension is in effect, all scale-out activities that are
              triggered by a scaling policy are suspended.
            * For ScheduledScalingSuspended, while a suspension is in effect,
              all scaling activities that involve scheduled actions are suspended.

            For more information, see Suspending and resuming scaling in the Application Auto Scaling User Guide.
            Defaults to None.

            * DynamicScalingInSuspended (bool, Optional):
                Whether scale in by a target tracking scaling policy or a step scaling policy is suspended. Set
                the value to true if you don't want Application Auto Scaling to remove capacity when a scaling
                policy is triggered. The default is false.

            * DynamicScalingOutSuspended (bool, Optional):
                Whether scale out by a target tracking scaling policy or a step scaling policy is suspended. Set
                the value to true if you don't want Application Auto Scaling to add capacity when a scaling
                policy is triggered. The default is false.

            * ScheduledScalingSuspended (bool, Optional):
                Whether scheduled scaling is suspended. Set the value to true if you don't want Application Auto
                Scaling to add or remove capacity by initiating scheduled actions. The default is false.

    Request Syntax:
        .. code-block:: sls

            [scalable_target_id]:
              aws.application_autoscaling.scalable_target.present:
                - name: 'string'
                - service_namespace: 'string'
                - scalable_dimension: 'string'
                - scaling_resource_id: 'string'
                - min_capacity: 'int'
                - max_capacity: 'int'
                - suspended_state:
                    DynamicScalingInSuspended: 'Boolean'
                    DynamicScalingOutSuspended: 'Boolean'
                    ScheduledScalingSuspended: 'Boolean'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            rds_scalable_target:
              aws.application_autoscaling.scalable_target.present:
                - name: rds_scalable_target
                - service_namespace: rds
                - scalable_dimension: rds:cluster:ReadReplicaCount
                - scaling_resource_id: idem-test-rds-aurora-table
                - min_capacity: 1
                - max_capacity: 3
                - suspended_state:
                    DynamicScalingInSuspended: False
                    DynamicScalingOutSuspended: False
                    ScheduledScalingSuspended: False

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    if resource_id:
        if not re.search(
            f"^({service_namespace})/({scaling_resource_id})/({scalable_dimension})$",
            resource_id,
        ):
            result["comment"] = (
                f"Incorrect aws.application_autoscaling.scalable_target resource_id: {resource_id}. Expected id {service_namespace}/{scaling_resource_id}/{scalable_dimension}",
            )
            result["result"] = False
            return result
        before = await hub.exec.aws.application_autoscaling.scalable_target.get(
            ctx=ctx,
            name=name,
            service_namespace=service_namespace,
            scaling_resource_ids=[scaling_resource_id],
            scalable_dimension=scalable_dimension,
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = await hub.tool.aws.application_autoscaling.scalable_target.update_scalable_target(
            ctx=ctx,
            name=name,
            before=result["old_state"],
            service_namespace=service_namespace,
            scaling_resource_id=scaling_resource_id,
            scalable_dimension=scalable_dimension,
            min_capacity=min_capacity,
            max_capacity=max_capacity,
            role_arn=role_arn,
            suspended_state=suspended_state,
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            for modified_param in update_ret["ret"]:
                plan_state[modified_param] = update_ret["ret"][modified_param]
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.application_autoscaling.scalable_target", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "service_namespace": service_namespace,
                    "scaling_resource_id": scaling_resource_id,
                    "scalable_dimension": scalable_dimension,
                    "min_capacity": min_capacity,
                    "max_capacity": max_capacity,
                    "role_arn": role_arn,
                    "suspended_state": suspended_state,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.application_autoscaling.scalable_target", name=name
            )
            return result
        ret = await hub.exec.boto3.client[
            "application-autoscaling"
        ].register_scalable_target(
            ctx,
            ServiceNamespace=service_namespace,
            ResourceId=scaling_resource_id,
            ScalableDimension=scalable_dimension,
            MinCapacity=min_capacity,
            MaxCapacity=max_capacity,
            RoleARN=role_arn,
            SuspendedState=suspended_state,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.application_autoscaling.scalable_target", name=name
        )
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.application_autoscaling.scalable_target.get(
                ctx=ctx,
                name=name,
                service_namespace=service_namespace,
                scaling_resource_ids=[scaling_resource_id],
                scalable_dimension=scalable_dimension,
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] += after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    scaling_resource_id: str,
    service_namespace: str,
    scalable_dimension: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Deregisters an Application Auto Scaling scalable target when you have finished using it.

    To see which resources have been registered, use DescribeScalableTargets.
    Deregistering a scalable target deletes the scaling policies and the scheduled actions that are associated with it.

    Args:
        name(str):
            An Idem name of the resource.

        service_namespace(str):
            The namespace of the Amazon Web Services service that provides the resource. For a resource
            provided by your own application or service, use custom-resource instead.

        scaling_resource_id(str):
            The identifier of the resource associated with the scalable target. This string consists of the
            resource type and unique identifier.

            * ECS service - The resource type is service and the unique identifier is the cluster name and service name.
              Example: service/default/sample-webapp.
            * Spot Fleet - The resource type is spot-fleet-request and the unique identifier is the Spot Fleet
              request ID. Example: spot-fleet-request/sfr-73fbd2ce-aa30-494c-8788-1cee4EXAMPLE.
            * EMR cluster - The resource type is instancegroup and the unique identifier is the cluster ID and instance
              group ID. Example: instancegroup/j-2EEZNYKUA1NTV/ig-1791Y4E1L8YI0.
            * AppStream 2.0 fleet - The resource type is fleet and the unique identifier is the fleet name.
              Example: fleet/sample-fleet.
            * DynamoDB table - The resource type is table and the unique identifier is the table name.
              Example: table/my-table.
            * DynamoDB global secondary index - The resource type is index and the
              unique identifier is the index name. Example: table/my-table/index/my-table-index.
            * Aurora DB cluster - The resource type is cluster and the unique identifier is the cluster name. Example:
              cluster:my-db-cluster.
            * SageMaker endpoint variant - The resource type is variant and the
              unique identifier is the resource ID. Example: endpoint/my-end-point/variant/KMeansClustering.
            * Custom resources are not supported with a resource type. This parameter must specify the
              OutputValue from the CloudFormation template stack used to access the resources. The unique
              identifier is defined by the service provider. More information is available in our GitHub
              repository.
            * Amazon Comprehend document classification endpoint - The resource type and unique
              identifier are specified using the endpoint ARN. Example: arn:aws:comprehend:us-
              west-2:123456789012:document-classifier-endpoint/EXAMPLE.
            * Amazon Comprehend entity recognizer endpoint -
              The resource type and unique identifier are specified using the endpoint ARN.
              Example: arn:aws:comprehend:us-west-2:123456789012:entity-recognizer-endpoint/EXAMPLE.
            * Lambda provisioned concurrency -
              The resource type is function and the unique identifier is the
              function name with a function version or alias name suffix that is not $LATEST. Example:
              function:my-function:prod or function:my-function:1.
            * Amazon Keyspaces table - The resource type is table and the unique identifier is the table name. Example:
              keyspace/mykeyspace/table/mytable.
            * Amazon MSK cluster -
              The resource type and unique identifier are specified using the cluster ARN. Example:
              arn:aws:kafka:us-east-1:123456789012:cluster/demo-cluster-1/6357e0b2-0e6a-4b86-a0b4-70df934c2e31-5.
            * Amazon ElastiCache replication group - The resource type is replication-group and the unique identifier
              is the replication group name. Example: replication-group/mycluster.
            * Neptune cluster - The resource type is cluster and the unique identifier is the cluster name. Example:
              cluster:mycluster.

        scalable_dimension(str):
            The scalable dimension associated with the scalable target. This string consists of the service
            namespace, resource type, and scaling property.

            * ecs:service:DesiredCount - The desired task count of an ECS service.
            * elasticmapreduce:instancegroup:InstanceCount - The instance count of an EMR Instance Group.
            * ec2:spot-fleet-request:TargetCapacity - The target capacity of a Spot Fleet.
            * appstream:fleet:DesiredCapacity - The desired capacity of an AppStream 2.0 fleet.
            * dynamodb:table:ReadCapacityUnits - The provisioned read capacity for a DynamoDB table.
            * dynamodb:table:WriteCapacityUnits - The provisioned write capacity for a DynamoDB table.
            * dynamodb:index:ReadCapacityUnits - The provisioned read capacity for a DynamoDB global secondary index.
            * dynamodb:index:WriteCapacityUnits - The provisioned write capacity for a DynamoDB global secondary index.
            * rds:cluster:ReadReplicaCount - The count of Aurora Replicas in an Aurora DB cluster.
              Available for Aurora MySQL-compatible edition and Aurora PostgreSQL- compatible edition.
            * sagemaker:variant:DesiredInstanceCount -
              The number of EC2 instances for an SageMaker model endpoint variant.
            * custom-resource:ResourceType:Property -
              The scalable dimension for a custom resource provided by your own application or service.
            * comprehend:document-classifier-endpoint:DesiredInferenceUnits - The number of inference units
              for an Amazon Comprehend document classification endpoint.
            * comprehend:entity-recognizer-endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend entity recognizer endpoint.
            * lambda:function:ProvisionedConcurrency - The provisioned concurrency for a Lambda function.
            * cassandra:table:ReadCapacityUnits - The provisioned read capacity for an Amazon Keyspaces table.
            * cassandra:table:WriteCapacityUnits - The provisioned write capacity for an Amazon Keyspaces table.
            * kafka:broker-storage:VolumeSize -
              The provisioned volume size (in GiB) for brokers in an Amazon MSK cluster.
            * elasticache:replication-group:NodeGroups -
              The number of node groups for an Amazon ElastiCache replication group.
            * elasticache:replication-group:Replicas -
              The number of replicas per node group for an Amazon ElastiCache replication group.
            * neptune:cluster:ReadReplicaCount - The count of read replicas in an Amazon Neptune DB cluster.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            rds_scalable_target:
              aws.application_autoscaling.scalable_target.absent:
                - name: rds_scalable_target
                - service_namespace: rds
                - scalable_dimension: rds:cluster:ReadReplicaCount
                - scaling_resource_id: idem-test-rds-aurora-table
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.application_autoscaling.scalable_target", name=name
        )
        return result
    if resource_id:
        if not re.search(
            f"^({service_namespace})/({scaling_resource_id})/({scalable_dimension})$",
            resource_id,
        ):
            result["comment"] = (
                f"Incorrect aws.application_autoscaling.scalable_target resource_id: {resource_id}. Expected id {service_namespace}/{scaling_resource_id}/{scalable_dimension}",
            )
            result["result"] = False
            return result
        before = await hub.exec.aws.application_autoscaling.scalable_target.get(
            ctx=ctx,
            name=name,
            service_namespace=service_namespace,
            scaling_resource_ids=[scaling_resource_id],
            scalable_dimension=scalable_dimension,
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.application_autoscaling.scalable_target", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.application_autoscaling.scalable_target", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client[
            "application-autoscaling"
        ].deregister_scalable_target(
            ctx,
            ServiceNamespace=service_namespace,
            ResourceId=scaling_resource_id,
            ScalableDimension=scalable_dimension,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = result[
            "comment"
        ] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.application_autoscaling.scalable_target", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Gets information about the scalable targets in the specified namespace. You can filter the results using
    ResourceIds and ScalableDimension.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.application_autoscaling.scalable_target
    """
    result = {}
    # service_name_spaces supported by AWS. loop through all the service name spaces
    # and list scaling targets of each service.
    service_name_spaces = [
        "ecs",
        "elasticmapreduce",
        "ec2",
        "appstream",
        "dynamodb",
        "rds",
        "sagemaker",
        "custom-resource",
        "comprehend",
        "lambda",
        "cassandra",
        "kafka",
        "elasticache",
        "neptune",
    ]
    for service_name_space in service_name_spaces:
        ret = await hub.exec.boto3.client[
            "application-autoscaling"
        ].describe_scalable_targets(ctx, ServiceNamespace=service_name_space)
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe scaling_targets for service name space {service_name_space}. {ret['comment']}"
            )
            continue
        for scaling_target in ret["ret"]["ScalableTargets"]:
            resource_id = f"{scaling_target.get('ServiceNamespace')}/{scaling_target.get('ResourceId')}/{scaling_target.get('ScalableDimension')}"
            resource_translated = hub.tool.aws.application_autoscaling.conversion_utils.convert_raw_scaling_target_to_present(
                ctx, raw_resource=scaling_target, idem_resource_name=resource_id
            )
            result[resource_id] = {
                "aws.application_autoscaling.scalable_target.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def convert_raw_budget_to_present(
    hub, ctx, account_id: str, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    resource_id = raw_resource.get("BudgetName")
    resource_parameters = OrderedDict(
        {
            "BudgetName": "budget_name",
            "BudgetLimit": "budget_limit",
            "CostFilters": "cost_filters",
            "CostTypes": "cost_types",
            "TimeUnit": "time_unit",
            "TimePeriod": "time_period",
            "CalculatedSpend": "calculated_spend",
            "ForecastedSpend": "forecasted_spend",
            "BudgetType": "budget_type",
            "LastUpdatedTime": "last_updated_time",
        }
    )

    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    ret_tag = await hub.exec.boto3.client.budgets.describe_notifications_for_budget(
        ctx, AccountId=account_id, BudgetName=resource_id
    )
    if ret_tag["result"]:
        notifications = ret_tag["ret"]["Notifications"]
        if notifications:
            resource_translated["notifications_with_subscribers"] = []
            for notification in notifications:
                ret_sub = await hub.exec.boto3.client.budgets.describe_subscribers_for_notification(
                    ctx,
                    AccountId=account_id,
                    BudgetName=resource_id,
                    Notification=notification,
                )
                notification_with_subscriber = {
                    "Notification": notification,
                    "Subscribers": ret_sub["ret"]["Subscribers"],
                }
                resource_translated["notifications_with_subscribers"].append(
                    notification_with_subscriber
                )

    return resource_translated


def convert_raw_budget_action_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    resource_id = raw_resource.get("ActionId")
    resource_parameters = OrderedDict(
        {
            "BudgetName": "budget_name",
            "NotificationType": "notification_type",
            "ActionType": "action_type",
            "ActionThreshold": "action_threshold",
            "Definition": "definition",
            "ExecutionRoleArn": "execution_role_arn",
            "ApprovalModel": "approval_model",
            "Subscribers": "subscribers",
        }
    )

    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



async def get_dhcp_vpc_association_details(hub, ctx, dhcp_id: str, vpc_id: str):
    result = dict(comment=(), result=True, ret=None)
    vpc_ret = await hub.exec.boto3.client.ec2.describe_vpcs(ctx, VpcIds=[vpc_id])
    if not vpc_ret["result"]:
        result["comment"] = vpc_ret["comment"]
        result["result"] = vpc_ret["result"]
        return result

    for vpc_dhcp in vpc_ret["ret"]["Vpcs"]:
        if vpc_dhcp.get("VpcId") == vpc_id and vpc_dhcp["DhcpOptionsId"] == dhcp_id:
            result["ret"] = vpc_dhcp
            return result
    result["result"] = False
    return result



"""Exec module for managing Amazon Neptune DB Instance."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}

RESOURCE_TYPE = "aws.neptune.db_instance"


async def get(
    hub, ctx, name: str, resource_id: str = None, filters: List = None
) -> Dict:
    """Fetch and use an un-managed neptune db instance as a data-source.

    Args:
        name(str):
            The name of the Idem state.
        resource_id(str, Optional):
            AWS Neptune DBInstanceIdentifier to identify the resource.
        filters(list, Optional):
            A filter that specifies one or more DB instances to describe.
            Supported filters are,

            * db-cluster(str, Optional):
                Accepts DB cluster identifiers and DB cluster Amazon Resource Names (ARNs). The results list will only include information about the DB instances associated with the DB clusters identified by these ARNs.
            * engine(str, Optional):
                Accepts an engine name (such as neptune ), and restricts the results list to DB instances created by that engine.

    .. note::
        If filters used matches more than one neptune db instance, idem will default to fetching the first result.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS Neptune DB Instance in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.neptune.db_instance.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.neptune.db_instance.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.neptune.db_instance.search_raw(
        ctx=ctx, name=name, resource_id=resource_id, filters=filters
    )
    if not ret["result"]:
        if "DBInstanceNotFoundFault" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type=RESOURCE_TYPE, name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBInstances"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        )
        return result

    resource = ret["ret"]["DBInstances"][0]
    if len(ret["ret"]["DBInstances"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type=RESOURCE_TYPE,
                resource_id=resource.get("DBInstanceIdentifier"),
            )
        )
    arn = resource["DBInstanceArn"]
    # get tags
    tags = await hub.tool.aws.neptune.tag.get_tags_for_resource(ctx, resource_arn=arn)
    if not tags["result"]:
        result["result"] = False
        result["comment"] = tags["comment"]
        return result
    tags = tags["ret"]
    result[
        "ret"
    ] = await hub.tool.aws.neptune.conversion_utils.convert_raw_db_instance_to_present(
        raw_resource=resource, idem_resource_name=name, tags=tags
    )
    return result


async def list_(hub, ctx, name, filters: List = None) -> Dict:
    """Use a list of un-managed neptune db instances as a data-source. Supply one of the inputs as the filter.

    Args:
        name(str):
            The name of idem state.
        filters(list, Optional):
            A filter that specifies one or more DB instances to describe.
            Supported filters are,

            * db-cluster(str, Optional):
                Accepts DB cluster identifiers and DB cluster Amazon Resource Names (ARNs). The results list will only include information about the DB instances associated with the DB clusters identified by these ARNs.
            * engine(str, Optional):
                Accepts an engine name (such as neptune ), and restricts the results list to DB instances created by that engine.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The AWS Neptune DB Instance in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.neptune.db_instance.list name="idem_name" filters=[db-cluster-id="", engine=""]

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.neptune.db_instance.get(
                    ctx, name=name, filters=[db-cluster-id="", engine=""]
                )
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.neptune.db_instance.search_raw(
        ctx=ctx, name=name, filters=filters
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["DBInstances"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type=RESOURCE_TYPE, name=name
            )
        )
        return result
    for db_instance in ret["ret"]["DBInstances"]:
        db_instance_identifier = db_instance.get("DBInstanceIdentifier")
        arn = db_instance["DBInstanceArn"]
        tags = await hub.tool.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=arn
        )
        tags = tags["ret"]
        result["ret"].append(
            await hub.tool.aws.neptune.conversion_utils.convert_raw_db_instance_to_present(
                raw_resource=db_instance,
                idem_resource_name=db_instance_identifier,
                tags=tags,
            )
        )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def search_raw(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict:
    """
    Fetch a load balancer from AWS. The return will be in the same format as what the boto3 api returns.

    Can't specify names and LoadBalancerArns in the same call. AWS API is failing with following exception.
        ['ClientError: An error occurred (ValidationError) when calling the 'DescribeLoadBalancers operation:
        Load balancer names and load balancer ARNs cannot be specified at the same time']
    Here, resource_id get higher priority in search than name i.e. if both name and resource_id are not None,
        search is done with resource_id than name.

    Args:
        name(Text): The name of the Idem state.
        resource_id(Text, Optional): AWS ELBv2 Load Balancer ARN to identify the resource.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    ret = result = dict(comment=[], ret=None, result=True)
    if resource_id:
        ret = await hub.exec.boto3.client.elbv2.describe_load_balancers(
            ctx,
            LoadBalancerArns=[resource_id],
        )
    elif name:
        ret = await hub.exec.boto3.client.elbv2.describe_load_balancers(
            ctx,
            Names=[name],
        )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    current_state: Dict[str, Any],
    input_map: Dict[str, Any],
    resource_id: str,
    plan_state: Dict[str, Any],
):
    r"""
    1. Associates the specified security groups with the specified Application Load Balancer. The specified security
       groups override the previously associated security groups. You can't specify a security group for a Network
       Load Balancer or Gateway Load Balancer.
    2. Enables the Availability Zones for the specified public subnets for the specified Application Load Balancer or
       Network Load Balancer. The specified subnets replace the previously enabled subnets.

       When you specify subnets for a Network Load Balancer, you must include all subnets that were enabled previously,
       with their existing configurations, plus any additional subnets.
    3. Modifies the specified attributes of the specified Application Load Balancer, Network Load Balancer, or
       Gateway Load Balancer.If any of the specified attributes can't be modified as requested, the call fails. Any
       existing attributes that you do not modify retain their current values.

    Args:
        current_state: response returned by describe on an AWS ELB load balancer
        input_map: a dictionary with newly passed values of params.
        resource_id: AWS ELB load balancer name.
        plan_state: idem --test state for update on AWS ElasticLoadBalancing Load Balancer.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": None}

    """
    result = dict(comment=[], result=True, ret=[])
    if not ctx.get("test", False):
        if input_map.get("Attributes"):
            attributes_update = (
                await hub.exec.boto3.client.elbv2.modify_load_balancer_attributes(
                    ctx,
                    LoadBalancerArn=resource_id,
                    Attributes=input_map.get("Attributes"),
                )
            )
            if not attributes_update["result"]:
                result["comment"] = list(attributes_update["comment"])
                result["result"] = False
                return result
            result["comment"].append("Modified Attributes.")
            result["ret"].append({"attributes": attributes_update["ret"]["Attributes"]})

        if input_map.get("SecurityGroups"):
            if not current_state.get("security_groups") or set(
                current_state.get("security_groups")
            ) != set(input_map.get("SecurityGroups")):
                update_security_groups = (
                    await hub.exec.boto3.client.elbv2.set_security_groups(
                        ctx,
                        LoadBalancerArn=resource_id,
                        SecurityGroups=input_map.get("SecurityGroups"),
                    )
                )
                if not update_security_groups["result"]:
                    result["comment"] += list(update_security_groups["comment"])
                    result["result"] = False
                    return result
                result["comment"].append("Updated SecurityGroups.")
                result["ret"].append(
                    {
                        "update_security_groups": update_security_groups["ret"][
                            "SecurityGroupIds"
                        ]
                    }
                )

        if input_map.get("Subnets"):
            if not current_state.get("subnets") or set(
                current_state.get("subnets")
            ) != set(input_map.get("Subnets")):
                update_subnets = await hub.exec.boto3.client.elbv2.set_subnets(
                    ctx,
                    LoadBalancerArn=resource_id,
                    Subnets=input_map.get("Subnets"),
                    IpAddressType=input_map.get("IpAddressType")
                    if input_map.get("Type") != "application"
                    else None,
                )
                if not update_subnets["result"]:
                    result["comment"] += list(update_subnets["comment"])
                    result["result"] = False
                    return result
                result["comment"].append("Updated Subnets.")
                result["ret"].append(
                    {"subnets_update": update_subnets["ret"]["AvailabilityZones"]}
                )

        if input_map.get("IpAddressType"):
            if not current_state.get("ip_address_type") or (
                current_state.get("ip_address_type") != input_map.get("IpAddressType")
            ):
                update_ip_address_type = (
                    await hub.exec.boto3.client.elbv2.set_ip_address_type(
                        ctx,
                        LoadBalancerArn=resource_id,
                        IpAddressType=input_map.get("IpAddressType"),
                    )
                )
                if not update_ip_address_type["result"]:
                    result["comment"] += list(update_ip_address_type["comment"])
                    result["result"] = False
                    return result
                result["comment"].append("Updated IpAddressType.")
                result["ret"].append(
                    {"ip_type_update": update_ip_address_type["ret"]["IpAddressType"]}
                )
    else:
        update_params = OrderedDict(
            {
                "name": input_map.get("LoadBalancerName"),
                "subnets": input_map.get("Subnets"),
                "subnet_mappings": input_map.get("SubnetMappings"),
                "security_groups": input_map.get("SecurityGroups"),
                "scheme": input_map.get("Scheme"),
                "tags": input_map.get("Tags"),
                "lb_type": input_map.get("Type"),
                "ip_address_type": input_map.get("IpAddressType"),
                "customer_owned_ipv4_pool": input_map.get("CustomerOwnedIpv4Pool"),
                "attributes": input_map.get("Attributes"),
                "resource_id": resource_id,
            }
        )
        for key, value in update_params.items():
            if value is not None:
                plan_state[key] = value

        result["ret"] = plan_state
    return result



import copy

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    parent_resource_id: str,
    http_method: str,
    input_type: str,
    resource_id: str = None,
    cache_namespace: str = None,
    cache_key_parameters: list = None,
    timeout_in_millis: int = None,
    passthrough_behavior: str = None,
    connection_id: str = None,
):
    """
    Creates a new API Gateway Integration or modifies an existing one.

    Args:
        hub:

        ctx:

        name(string):
            An idem name of the resource.

        rest_api_id(string):
            AWS rest_api id of the associated RestApi.

        parent_resource_id(string):
            The parent resource's id.

        http_method(string):
            String that specifies the method request's HTTP method type.

        input_type(string):
            Specifies a put integration input's type.

        resource_id(string, optional):
            Idem Resource id, formatted as: rest_api_id-parent_resource_id-http_method. Defaults to None.

        cache_namespace(string, optional):
            Specifies a group of related cached parameters. By default, API Gateway uses
                the resource ID as the cacheNamespace . You can specify the same cacheNamespace across resources to return
                the same cached data for requests to different resources.

        cache_key_parameters(list, optional):
            A list of request parameters whose values API Gateway caches. To be valid
                values for cacheKeyParameters , these parameters must also be specified for Method requestParameters.

        timeout_in_millis(int, optional):
            Custom timeout between 50 and 29,000 milliseconds. The default value is 29,000 milliseconds or 29 seconds.

        passthrough_behavior:
            Custom timeout between 50 and 29,000 milliseconds. The default value is
                29,000 milliseconds or 29 seconds.

        connection_id(string, optional):
            The ID of the VpcLink used for the integration. Specify this value only if you specify VPC_LINK as the
                connection type.

    Request Syntax:
        [idem_test_aws_apigateway_integration]:
          aws.apigateway.integration.present:
            - name: 'string'
            - rest_api_id: 'string'
            - parent_resource_id: 'string'
            - http_method: 'string'
            - input_type: 'string'

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.integration.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_parameters = {
            "cache_namespace": cache_namespace,
            "cache_key_parameters": cache_key_parameters,
            "input_type": input_type,
            "timeout_in_millis": timeout_in_millis,
            "passthrough_behavior": passthrough_behavior,
            "connection_id": connection_id,
        }

        update_ret = await hub.tool.aws.apigateway.integration.update(
            ctx,
            old_state=result["old_state"],
            updateable_parameters=update_parameters,
        )

        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        result["comment"] = result["comment"] + update_ret["comment"]
        resource_updated = bool(update_ret["ret"])

        if resource_updated and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.integration", name=name
                )
                return result

    else:
        resource_id = f"{rest_api_id}-{parent_resource_id}-{http_method}"

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "rest_api_id": rest_api_id,
                    "resource_id": resource_id,
                    "parent_resource_id": parent_resource_id,
                    "http_method": http_method,
                    "passthrough_behavior": passthrough_behavior,
                    "timeout_in_millis": timeout_in_millis,
                    "cache_namespace": cache_namespace,
                    "cache_key_parameters": cache_key_parameters,
                    "connection_id": connection_id,
                    "input_type": input_type,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.integration", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.put_integration(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            type=input_type,
            cacheNamespace=cache_namespace,
            cacheKeyParameters=cache_key_parameters,
            timeoutInMillis=timeout_in_millis,
            passthroughBehavior=passthrough_behavior,
            connectionId=connection_id,
        )

        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.integration", name=name
        )
    if (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.integration.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result

        resource_translated = after["ret"]
        result["new_state"] = resource_translated

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
):
    r"""

    Deletes an API Gateway Integration.

    Args:
        hub:

        ctx:

        name(string):
            An Idem name of the resource.

        resource_id(string, optional):
            Idem Resource id, formatted as: rest_api_id-parent_resource_id-http_method. Defaults to None.
                Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

        idem_test_aws_apigateway_integration:
          aws.apigateway.integration.absent:
            - name: value
            - resource_id: value

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.integration.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result

    else:
        rest_api_id, parent_resource_id, http_method = resource_id.split("-")
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_integration(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.integration", name=name
        )
        return result


async def describe(hub, ctx):
    r"""

    Describe the API Gateway Integrations.

    Returns a list of apigateway.integration descriptions

    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.integration

    """
    result = {}

    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    for rest_api in get_rest_apis_ret["ret"]["items"]:
        rest_api_id = rest_api.get("id")
        get_resources_ret = await hub.exec.boto3.client.apigateway.get_resources(
            ctx, restApiId=rest_api_id
        )
        if not get_resources_ret["result"]:
            f"Could not get Resources {get_resources_ret['comment']}. Will skip this Resource."
            continue

        if get_resources_ret["ret"]["items"] is not None:
            for resource in get_resources_ret["ret"]["items"]:
                parent_resource_id = resource.get("id")
                if resource.get("resourceMethods") is not None:
                    for resource_method in resource.get("resourceMethods"):
                        method_resource_id = (
                            f"{rest_api_id}-{parent_resource_id}-{resource_method}"
                        )

                        integration = await hub.exec.aws.apigateway.integration.get(
                            ctx,
                            resource_id=method_resource_id,
                        )

                        if not integration["ret"]:
                            hub.log.debug(
                                f"Could not get Integration {integration['comment']}. Will skip this Integration."
                            )
                            continue

                        resource_translated = integration["ret"]
                        result[resource_translated["resource_id"]] = {
                            "aws.apigateway.integration.present": [
                                {parameter_key: parameter_value}
                                for parameter_key, parameter_value in resource_translated.items()
                            ]
                        }

        return result



from typing import Dict

ARN_PREFIX = "arn"
ARN_DELIMITER = ":"


def build(
    hub,
    *,
    partition: str = "aws",
    service: str,
    region: str = None,
    account_id: str = None,
    resource: str,
) -> str:
    """
    Creates an AWS ARN from parameters.

    Args:
        hub: The redistributed pop central hub.
        partition(str, optional): The partition that the resource is in. For standard AWS regions, the partition is "aws" (default).
                If you have resources in other partitions, the partition is "aws-partitionname". For example, the partition for resources
                in the China (Beijing) region is "aws-cn".
        service(str): The service namespace that identifies the AWS product (for example, Amazon S3, IAM, or Amazon RDS). For a list of
                namespaces, see http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-aws-service-namespaces.
        region(str, optional): The region the resource resides in. Note that the ARNs for some resources do not require a region, so this
                component might be omitted.
        account_id(str, optional): The ID of the AWS account that owns the resource, without the hyphens. For example, 123456789012. Note that the
                ARNs for some resources don't require an account number, so this component might be omitted.
        resource(str): The content of this part of the ARN varies by service. It often includes an indicator of the type of resource —
                for example, an IAM user or Amazon RDS database - followed by a slash (/) or a colon (:), followed by the
                resource name itself. Some services allows paths for resource names, as described in
                http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#arns-paths.

    Returns:
        The ARN (str)
    """
    return (
        ARN_PREFIX
        + ARN_DELIMITER
        + partition
        + ARN_DELIMITER
        + service
        + ARN_DELIMITER
        + str(region or "")
        + ARN_DELIMITER
        + str(account_id or "")
        + ARN_DELIMITER
        + str(resource or "")
    )


def parse(hub, arn: str) -> Dict:
    """
    Parses the provided AWS ARN and returns its components.

    Args:
        hub: The redistributed pop central hub.
        arn(string): The AWS ARN.

    Returns:
        The ARN parts (Dict):
          partition: The partition that the resource is in.
          service: The service namespace that identifies the AWS product.
          region: The region the resource resides in.
          account_id: The ID of the AWS account that owns the resource, without the hyphens.
          resource: The resource.
    """
    parts = arn.split(ARN_DELIMITER)

    if len(parts) < 6:
        raise TypeError(f"Invalid AWS ARN {arn}")

    arn_parts = {
        "partition": parts[1],
        "service": parts[2],
        "region": parts[3],
        "account_id": parts[4],
        "resource": ":".join(parts[5:]),
    }

    return arn_parts


def get_resource_name(hub, arn: str) -> str:
    r"""
    Get the resource name from a provided AWS ARN.

    The resource part of an AWS ARN may include a name for the resource:
        arn:partition:service:region:account-id:resource-type:resource-name:qualifier

    Args:
        hub: The redistributed pop central hub.
        arn(string): The AWS ARN.

    Returns:
        str: The resource name.
    """
    parts = arn.split(ARN_DELIMITER)
    if len(parts) < 7 or not parts[6]:
        return ""
    return parts[6]


def get_qualifier(hub, arn: str) -> str:
    """
    Get the qualifier from a provided AWS ARN.

    The resource part of an AWS ARN may include a qualifier for the resource such as a version:
        arn:partition:service:region:account-id:resource-type:resource-name:qualifier

    Args:
        hub: The redistributed pop central hub.
        arn(string): The AWS ARN.

    Returns:
        str: The qualifier.
    """
    parts = arn.split(ARN_DELIMITER)
    if len(parts) < 8 or not parts[7]:
        return ""
    return parts[7]



from typing import List

# Modify this attribute last
__update_order__ = -1


async def apply(
    hub, ctx, resource, *, old_value, new_value, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    if new_value is True:
        # Start the instance up again
        ret = await hub.exec.boto3.client.ec2.start_instances(InstanceIds=[resource.id])
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False
        await hub.tool.boto3.resource.exec(resource, "wait_until_running")
    elif new_value is False:
        # Stopping the instance
        ret = await hub.exec.boto3.client.ec2.stop_instances(
            InstanceIds=[resource.id], Hibernate=False, Force=False
        )
        await hub.tool.boto3.resource.exec(resource, "wait_until_stopped")

        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            return False

    return True



"""

hub.exec.boto3.client.ec2.associate_dhcp_options
"""
import copy
import re
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    dhcp_id: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Associates a set of DHCP options (that you've previously created) with the specified VPC, or associates no DHCP
    options with the VPC.

    After you associate the options with the VPC, any existing instances and all new instances that you launch in that
    VPC use the options. You don't need to restart or relaunch the instances. They automatically pick up the changes
    within a few hours, depending on how frequently the instance renews its DHCP lease. You can explicitly renew the
    lease using the operating system on the instance.

    Args:
        name(str):
            An Idem name to identify the dhcp option resource.

        vpc_id(str):
            The list of AWS VPC ID that needs to be associated with dhcp options set.

        dhcp_id(str):
            AWS DHCP Option Set ID.

        resource_id(str, Optional):
            Vpc ID and DHCP Options ID with a separator '-'. Format: [dhcp_id]-[vpc_id]

    Request Syntax:
        .. code-block:: sls

            [dhcp_option-name]:
              aws.ec2.dhcp_option_association.present:
                - resource_id: 'string'
                - vpc_id: 'string'
                - dhcp_id: 'string'
                - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            my-dhcp-option:
              aws.ec2.dhcp_option_association.present:
                - resource_id: dhcp-8dh27j-vpc-76db75b8
                - vpc_id: vpc-76db75b8
                - dhcp_id: dhcp-8dh27j
                - name: my-dhcp-option
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if resource_id:
        if not re.search(f"^({dhcp_id})-({vpc_id})$", resource_id):
            result["comment"] = (
                f"Incorrect aws.ec2.dhcp_option_association resource_id: {resource_id}. Expected id {dhcp_id}-{vpc_id}",
            )
            result["result"] = False
            return result

    resource = hub.tool.boto3.resource.create(
        ctx,
        "ec2",
        "DhcpOptions",
        dhcp_id,
    )
    before = await hub.tool.boto3.resource.describe(resource)
    if before:
        ret = await hub.tool.aws.ec2.dhcp_option_association_utils.get_dhcp_vpc_association_details(
            ctx, dhcp_id=dhcp_id, vpc_id=vpc_id
        )
        if ret["result"]:
            result[
                "old_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_association_to_present(
                dhcp_id=dhcp_id,
                idem_resource_name=name,
                vpc_id=vpc_id,
            )
            result["comment"] = (
                f"aws.ec2.dhcp_option_association '{name}' for vpc_id '{vpc_id}' already exists",
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "vpc_id": vpc_id,
                    "dhcp_id": dhcp_id,
                    "resource_id": f"{dhcp_id}-{vpc_id}",
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.dhcp_option_association", name=name
            )
            return result
        ret = await hub.exec.boto3.client.ec2.associate_dhcp_options(
            ctx,
            DhcpOptionsId=dhcp_id,
            VpcId=vpc_id,
        )
        if not ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + ret["comment"]
            return result
        result["comment"] = result[
            "comment"
        ] + hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.dhcp_option_association", name=name
        )
    else:
        result["result"] = False
        result["comment"] = (
            f"Incorrect aws.ec2.dhcp_option_association id: '{dhcp_id}'. Couldn't find DHCP Options with given id.",
        )
        return result
    try:
        convert_ret = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_association_to_present(
                dhcp_id=dhcp_id,
                idem_resource_name=name,
                vpc_id=vpc_id,
            )
        )
        result["new_state"] = convert_ret
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    dhcp_id: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Deletes/disassociates association of DHCP options with VPC.

    Args:
        name(str):
            An Idem name to identify the dhcp option resource.

        vpc_id (str):
            AWS VPC ID that needs to be associated with dhcp options set.

        dhcp_id (str):
            AWS DHCP Option Set ID.

        resource_id(str, Optional):
            Vpc ID and DHCP Options ID with a separator '-'. Format: [dhcp_id]-[vpc_id]

    Request Syntax:
        .. code-block:: sls

            [dhcp_option-name]:
              aws.ec2.dhcp_option_association.absent:
                - resource_id: 'string'
                - vpc_id: 'string'
                - dhcp_id: 'string'
                - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            my-dhcp-option:
              aws.ec2.dhcp_option_association.absent:
                - resource_id: dhcp-8dh27j-vpc-76db75b8
                - vpc_id: vpc-76db75b8
                - dhcp_id: dhcp-8dh27j
                - name: my-dhcp-option
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if resource_id:
        if not re.search(f"^({dhcp_id})-({vpc_id})$", resource_id):
            result[
                "comment"
            ] = f"Incorrect aws.ec2.dhcp_option_association resource_id: {resource_id}. Expected id {dhcp_id}-{vpc_id}"
            result["result"] = False
            return result
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "DhcpOptions", dhcp_id)
    before = await hub.tool.boto3.resource.describe(resource)
    if not before:
        result["result"] = False
        result["comment"] = (
            f"Incorrect aws.ec2.dhcp_option_association id: '{dhcp_id}'. Couldn't find DHCP Options with given id.",
        )
        return result
    else:
        ret = await hub.tool.aws.ec2.dhcp_option_association_utils.get_dhcp_vpc_association_details(
            ctx, dhcp_id=dhcp_id, vpc_id=vpc_id
        )
        if not ret["result"]:
            result["comment"] = (
                f"aws.ec2.dhcp_option_association '{name}' for vpc_id '{vpc_id}' is already absent!",
            )
            return result
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_association_to_present(
            dhcp_id=dhcp_id,
            idem_resource_name=name,
            vpc_id=vpc_id,
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.ec2.dhcp_option_association", name=name
            )
            return result
        update_ret = await hub.exec.boto3.client.ec2.associate_dhcp_options(
            ctx,
            DhcpOptionsId="default",
            VpcId=vpc_id,
        )
        result["result"] = update_ret["result"]
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.dhcp_option_association", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Lists out all DHCP Options to VPC associations.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.dhcp_option_association
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_dhcp_options(ctx)
    vpc_ret = await hub.exec.boto3.client.ec2.describe_vpcs(ctx)

    if not (ret["result"] and vpc_ret["result"]):
        hub.log.debug(f"Could not describe dhcp_options association {ret['comment']}")
        return result

    for resource in ret["ret"]["DhcpOptions"]:
        resource_id = resource.get("DhcpOptionsId")
        for vpc_dhcp in vpc_ret["ret"]["Vpcs"]:
            if vpc_dhcp["DhcpOptionsId"] == resource_id:
                resource_converted = hub.tool.aws.ec2.conversion_utils.convert_raw_dhcp_association_to_present(
                    dhcp_id=resource_id,
                    idem_resource_name=resource_id,
                    vpc_id=vpc_dhcp["VpcId"],
                )
                result[f"{resource_id}-{vpc_dhcp['VpcId']}"] = {
                    "aws.ec2.dhcp_option_association.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_converted.items()
                    ]
                }
    return result



import base64


def decode(hub, data: bytes) -> bytes:
    if not isinstance(data, bytes):
        data = str(data).encode("utf-8")
    decoded_data = base64.b64decode(data)
    return decoded_data


def encode(hub, data: bytes) -> str:
    if not isinstance(data, bytes):
        data = str(data).encode("utf-8")
    b64_data_bytes = base64.b64encode(data)
    b64_data_str = b64_data_bytes.decode("utf-8")
    return b64_data_str



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str = None,
    resource_id: str = None,
    all_availability_zones: bool = None,
    filters: List = None,
) -> Dict:
    """
    Returns the list of AWS Availability Zones, Local Zones, and Wavelength Zones that are available
    within the region. The return will be in the same format as what the boto3 api returns.

    Args:
        hub: The redistributed pop central hub.
        ctx: The context.
        name(string, optional): The name of the AWS Availability Zone, Local Zone, or Wavelength Zone.
        resource_id(string, optional): The id of the of the AWS Availability Zone, Local Zone, or Wavelength Zone.
        all_availability_zones(bool, optional): Indicate whether to include all Availability Zones, Local Zones, and Wavelength Zones regardless of the opt-in status.
        filters(list, optional): One or more filters. For example: {"Name": "state", "Values": ["available"]},
          A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_availability_zones

    .. code-block:: bash

        $ idem exec aws.ec2.availability_zones.get name=us-west-2b
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.tool.aws.ec2.availability_zones.search_raw(
        ctx=ctx,
        zone_names=[name] if name else None,
        zone_ids=[resource_id] if resource_id else None,
        all_availability_zones=all_availability_zones,
        filters=filters,
    )
    if not ret["result"]:
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["AvailabilityZones"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.availability_zones",
                name=name,
            )
        )
        return result

    resource = ret["ret"]["AvailabilityZones"][0]
    if len(ret["ret"]["AvailabilityZones"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.ec2.availability_zones",
                resource_id=resource.get("ZoneId"),
            )
        )
    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_availability_zone_to_snake_case(
        ctx, raw_resource=resource
    )

    return result


async def list_(
    hub, ctx, all_availability_zones: bool = None, filters: List = None
) -> Dict:
    """
    Returns the list of AWS Availability Zones, Local Zones, and Wavelength Zones that are available
    within the region. The return will be in the same format as what the boto3 api returns.

    Args:
        hub: The redistributed pop central hub.
        ctx: The context.
        all_availability_zones(bool, optional): Indicate whether to include all Availability Zones, Local Zones, and Wavelength Zones regardless of the opt-in status.
        filters(list, optional): One or more filters. For example: {"Name": "state", "Values": ["available"]},
          A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_availability_zones

    .. code-block:: bash

        $ idem exec aws.ec2.availability_zones.list all_availability_zones=true
    """
    result = dict(comment=[], ret=[], result=True)

    ret = await hub.tool.aws.ec2.availability_zones.search_raw(
        ctx=ctx,
        all_availability_zones=all_availability_zones,
        filters=filters,
    )
    if not ret["result"]:
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["AvailabilityZones"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.availability_zones"
            )
        )
        return result

    for az in ret["ret"]["AvailabilityZones"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_availability_zone_to_snake_case(
                ctx, raw_resource=az
            )
        )

    return result



from typing import Any
from typing import Dict


async def is_hosted_zone_associated_to_vpc(
    hub, ctx, zone_id: str, vpc_id: str, vpc_region: str
) -> Dict[str, Any]:
    """
    Check if a vpc is attached to a hosted zone

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        zone_id: The ID of the hosted zone to be associated with.
        vpc_id: The ID of the vpc you want to associate.
        vpc_region: The AWS region where the vpc belongs to.

    Returns:
        {"result": True|False, "comment": (f"A message",), "ret": None}
    """
    result = dict(comment=(), result=True, ret=None)
    ret_list = await hub.exec.boto3.client.route53.list_hosted_zones_by_vpc(
        ctx, VPCId=vpc_id, VPCRegion=vpc_region
    )
    if ret_list["result"]:
        if ret_list["ret"].get("HostedZoneSummaries"):
            associated_zone_summaries = ret_list["ret"].get("HostedZoneSummaries")
            hosted_zone_list = [
                hosted_zone_summary.get("HostedZoneId")
                for hosted_zone_summary in associated_zone_summaries
            ]
            if zone_id in hosted_zone_list:
                result["ret"] = {"HostedZone": zone_id}
    else:
        result["result"] = False
        result["comment"] = ret_list["comment"]
    return result



import json
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def convert_raw_db_instance_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to idem neptune db_instance present state.
    Following parameters are part of present but are not available as part of describe:
    - master_user_password
    - tde_credential_password
    Following parameter is not part of present and is ignored from describe:
    - db_instance_status

    Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert
        idem_resource_name (str, Optional): The idem name of the resource
        tags (Dict, Optional): The tags of the resource. Defaults to None.

    Returns: Valid idem state for neptune db_instance of type Dict['string', Any]
    """
    resource_id = raw_resource.get("DBInstanceIdentifier")
    raw_resource["Tags"] = dict(tags) if tags else None
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    resource_parameters = OrderedDict(
        {
            "DBInstanceIdentifier": "db_instance_identifier",
            "DBInstanceClass": "db_instance_class",
            "Engine": "engine",
            "DBName": "db_name",
            "AllocatedStorage": "allocated_storage",
            "MasterUsername": "master_username",
            "DBSecurityGroups": "db_security_groups",
            "AvailabilityZone": "availability_zone",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "BackupRetentionPeriod": "backup_retention_period",
            "PreferredBackupWindow": "preferred_backup_window",
            "DBInstancePort": "port",
            "MultiAZ": "multi_az",
            "EngineVersion": "engine_version",
            "AutoMinorVersionUpgrade": "auto_minor_version_upgrade",
            "LicenseModel": "license_model",
            "Iops": "iops",
            "CharacterSetName": "character_set_name",
            "PubliclyAccessible": "publicly_accessible",
            "Tags": "tags",
            "DBClusterIdentifier": "db_cluster_identifier",
            "StorageType": "storage_type",
            "TdeCredentialArn": "tde_credential_arn",
            "StorageEncrypted": "storage_encrypted",
            "KmsKeyId": "kms_key_id",
            "CopyTagsToSnapshot": "copy_tags_to_snapshot",
            "MonitoringInterval": "monitoring_interval",
            "MonitoringRoleArn": "monitoring_role_arn",
            "PromotionTier": "promotion_tier",
            "Timezone": "timezone",
            "IAMDatabaseAuthenticationEnabled": "enable_iam_database_authentication",
            "PerformanceInsightsEnabled": "enable_performance_insights",
            "PerformanceInsightsKMSKeyId": "performance_insights_kms_key_id",
            "EnabledCloudwatchLogsExports": "enable_cloudwatch_logs_exports",
            "DBInstanceArn": "db_instance_arn",
            "DeletionProtection": "deletion_protection",
        }
    )
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    # lambda to extract values from dict containing a list of dict
    list_processor = lambda r, a, b: [e.get(b) for e in r.get(a)]
    get_first = lambda l: next((e for e in l if e is not None), None)

    # process the VpcSecurityGroups which is a list of object to map it as list of vpc security group id strings
    if raw_resource.get("VpcSecurityGroups") is not None:
        resource_translated["vpc_security_group_ids"] = list_processor(
            r=raw_resource, a="VpcSecurityGroups", b="VpcSecurityGroupId"
        )

    # process db parameter group name
    if raw_resource.get("DBParameterGroups") is not None:
        # even though this is returned as list, we are only allowed to associate one DB parameter group per instance, so below should execute once
        db_param_grp_name = get_first(
            list_processor(
                r=raw_resource, a="DBParameterGroups", b="DBParameterGroupName"
            )
        )
        if db_param_grp_name:
            resource_translated["db_parameter_group_name"] = db_param_grp_name

    # process for option_group_name
    if raw_resource.get("OptionGroupMemberships") is not None:
        option_group_name = get_first(
            list_processor(
                r=raw_resource, a="OptionGroupMemberships", b="OptionGroupName"
            )
        )
        if option_group_name:
            resource_translated["option_group_name"] = option_group_name

    # process domain and domain_iam_role_name
    if raw_resource.get("DomainMemberships") is not None:
        domain = get_first(
            list_processor(r=raw_resource, a="DomainMemberships", b="Domain")
        )
        if domain:
            resource_translated["domain"] = domain

        domain_iam_role_name = get_first(
            list_processor(r=raw_resource, a="DomainMemberships", b="DomainIAMRoleName")
        )
        if domain_iam_role_name:
            resource_translated["domain_iam_role_name"] = domain_iam_role_name

    # process db_subnet_group_name
    if raw_resource.get("DBSubnetGroup") is not None:
        db_subnet_group_name = raw_resource["DBSubnetGroup"].get("DBSubnetGroupName")
        if db_subnet_group_name:
            resource_translated["db_subnet_group_name"] = db_subnet_group_name
    return resource_translated


async def convert_raw_db_cluster_parameter_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    parameters: List[Dict[str, Any]] = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem db_cluster_parameter_group present state

     Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert
        idem_resource_name (str, Optional): The idem of the idem resource
        parameters (List, Optional): The DBClusterParameterGroup Parameters.
        tags (List, Optional): The tags of the resource. Defaults to None.

    Returns: Valid idem state for db_cluster_parameter_group of type Dict['string', Any]
    """
    resource_id = raw_resource.get("DBClusterParameterGroupName")
    raw_resource["Tags"] = tags
    raw_resource["Parameters"] = parameters
    resource_parameters = OrderedDict(
        {
            "DBParameterGroupFamily": "db_parameter_group_family",
            "Description": "description",
            "Tags": "tags",
            "Parameters": "parameters",
            "DBClusterParameterGroupArn": "db_cluster_parameter_group_arn",
        }
    )
    resource_translated = {
        "name": idem_resource_name if idem_resource_name else resource_id,
        "resource_id": resource_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = json.loads(
                json.dumps(raw_resource.get(parameter_raw))
            )
    return resource_translated



import copy
import re
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    zone_id: str,
    vpc_id: str,
    vpc_region: str = None,
    resource_id: str = None,
    comment: str = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Associates the specified vpc to the specified hosted zone.

    Args:
        hub:
        ctx:
        name(Text): An Idem name of the AWS hosted zone vpc association.
        zone_id(Text): The id of hosted zone to associate to a vpc
        vpc_id(Text): The id of the vpc to associate to a hosted zone.
        vpc_region(Text, optional): The AWS region where the vpc belongs to.
                                    If the vpc_region is not specified, AWS region from credentials file will be used.
        resource_id(Text, optional): The identifier for this object, combination of hosted_zone id, vpc id and vpc region separated by a separator ':'
        comment(Text, optional): A comment about the association request.
        timeout(Dict, optional): Timeout configuration for create/update of AWS DB Cluster.
            * create (Dict) -- Timeout configuration for creating DB Cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
        timeout(Dict, optional): Timeout configuration for hosted zone and vpc association.
            * create (Dict) -- Timeout configuration for creating nat gateway
                 * delay (int) -- The amount of time in seconds to wait between attempts. Defaults to 15.
                * max_attempts(int) -- Customized timeout configuration containing delay and max attempts. Defaults to 40.

    Request Syntax:
        [zone_id:vpc_id:vpc_region]:
          aws.route53.hosted_zone_association.present:
          - resource_id: 'string'
          - zone_id: 'string'
          - vpc_id: 'string'
          - vpc_region: 'string'
          - comment: 'string'
          - timeout:
              create
                delay: 'integer'
                max_attempts: 'integer'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            Z09756241DYDBTZK8J11E:vpc-9aacf0f2:eu-central-1:
              aws.route53.hosted_zone_association.present:
                - zone_id: Z09756241DYDBTZK8J11E
                - vpc_id: vpc-9aacf0f2
                - vpc_region: eu-central-1
                - resource_id: Z09756241DYDBTZK8J11E:vpc-9aacf0f2:eu-central-1
                - timeout:
                    create:
                        delay: 15
                        max_attempts: 40

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        if (
            not re.findall(":", resource_id)
            or not len(re.findall(":", resource_id)) == 2
        ):
            result[
                "comment"
            ] = f"Incorrect aws.route53.hosted_zone_association resource_id: {resource_id}. Expected id <zoneId>:<vpcId>:<vpc region>"
            result["result"] = False
            return result
        resource = resource_id.split(":")
        vpc_region = resource[2]

    if vpc_region is None:
        vpc_region = ctx["acct"].get("region_name")
    if not (vpc_id == "resource_id_known_after_present"):
        before = await hub.exec.aws.route53.hosted_zone_association.is_hosted_zone_associated_to_vpc(
            ctx, zone_id=zone_id, vpc_id=vpc_id, vpc_region=vpc_region
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
    if before and before["ret"]:
        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_zone_association_to_present(
            hosted_zone_id=zone_id,
            vpc_id=vpc_id,
            vpc_region=vpc_region,
            idem_resource_name=name,
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (
            f"aws.route53.hosted_zone_association '{name}' association already exists",
        )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "zone_id": zone_id,
                    "vpc_id": vpc_id,
                    "vpc_region": vpc_region,
                    "comment": comment,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.route53.hosted_zone_association", name=name
            )
            return result
        associate_ret = (
            await hub.exec.boto3.client.route53.associate_vpc_with_hosted_zone(
                ctx,
                HostedZoneId=zone_id,
                VPC={"VPCId": vpc_id, "VPCRegion": vpc_region},
                Comment=comment,
            )
        )
        result["result"] = associate_ret["result"]
        if not result["result"]:
            result["comment"] = associate_ret["comment"]
            return result

        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("create") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "route53",
                "resource_record_sets_changed",
                Id=associate_ret["ret"]["ChangeInfo"]["Id"],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

        resource_id = f"{zone_id}:{vpc_id}:{vpc_region}"
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.route53.hosted_zone_association", name=name
        )

        result[
            "new_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_zone_association_to_present(
            hosted_zone_id=zone_id,
            vpc_id=vpc_id,
            vpc_region=vpc_region,
            idem_resource_name=resource_id,
            comment=associate_ret["ret"]["ChangeInfo"].get("Comment"),
        )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    comment: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    """

    Removes the specified managed policy from the specified role.

    Args:
        name(Text): An Idem name of the AWS hosted zone vpc association.
        resource_id(Text, optional): The id of the hosted zone and vpc association. Idem automatically considers this resource being absent
         if this field is not specified.
        comment(Text, optional): A comment about the disassociation request.
        timeout(Dict, optional): Timeout configuration for hosted zone and vpc association.
            * delete (Dict) -- Timeout configuration for creating nat gateway
                    * delay (int) -- The amount of time in seconds to wait between attempts. Defaults to 15.
                    * max_attempts(int) -- Customized timeout configuration containing delay and max attempts. Defaults to 40.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            Z09756241DYDBTZK8J11E:vpc-9aacf0f2:eu-central-1:
              aws.route53.hosted_zone_association.absent:
                - name: Z09756241DYDBTZK8J11E:vpc-9aacf0f2:eu-central-1
                - resource_id: Z09756241DYDBTZK8J11E:vpc-9aacf0f2:eu-central-1
                - timeout:
                    delete:
                        delay: 15
                        max_attempts: 40
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.hosted_zone_association", name=name
        )
        return result
    if not re.findall(":", resource_id) or not len(re.findall(":", resource_id)) == 2:
        result[
            "comment"
        ] = f"Incorrect aws.route53.hosted_zone_association resource_id: {resource_id}. Expected id <zoneId>:<vpcId>:<vpc region>"
        result["result"] = False
        return result
    resource = resource_id.split(":")
    zone_id = resource[0]
    vpc_id = resource[1]
    vpc_region = resource[2]

    if not vpc_region:
        vpc_region = ctx["acct"].get("region_name")

    before_ret = await hub.exec.aws.route53.hosted_zone_association.is_hosted_zone_associated_to_vpc(
        ctx, zone_id=zone_id, vpc_id=vpc_id, vpc_region=vpc_region
    )

    if not before_ret["result"]:
        if "InvalidInput" in before_ret["comment"][0]:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.route53.hosted_zone_association", name=name
            )
        else:
            result["comment"] = before_ret["comment"]
            result["result"] = False
        return result
    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.hosted_zone_association", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_zone_association_to_present(
            hosted_zone_id=zone_id,
            vpc_id=vpc_id,
            vpc_region=vpc_region,
            idem_resource_name=name,
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.route53.hosted_zone_association", name=name
            )
            return result
        disassociate_ret = (
            await hub.exec.boto3.client.route53.disassociate_vpc_from_hosted_zone(
                ctx,
                HostedZoneId=zone_id,
                VPC={"VPCRegion": vpc_region, "VPCId": vpc_id},
                Comment=comment,
            )
        )
        result["result"] = disassociate_ret["result"]
        if not result["result"]:
            result["comment"] = disassociate_ret["comment"]
            return result

        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "route53",
                "resource_record_sets_changed",
                Id=disassociate_ret["ret"]["ChangeInfo"]["Id"],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.route53.hosted_zone_association", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the associated vpc's with respective hosted zones. If there are no vpc associated with the specified hosted
    zone, the operation returns an empty dict.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.route53.hosted_zone_association

    """

    result = {}
    ret = await hub.exec.boto3.client.route53.list_hosted_zones(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.route53.hosted_zone, {ret['comment']}")
        return {}

    for hosted_zone in ret["ret"]["HostedZones"]:

        hosted_zone_id = hosted_zone.get("Id")

        hosted_zone_details = await hub.exec.boto3.client.route53.get_hosted_zone(
            ctx, Id=hosted_zone_id
        )
        if hosted_zone_details and hosted_zone_details["ret"].get("VPCs"):
            associated_vpcs = hosted_zone_details["ret"].get("VPCs")

            for vpc in associated_vpcs:
                vpc_id = vpc.get("VPCId")
                vpc_region = vpc.get("VPCRegion")
                hosted_zone_id = hosted_zone_id.split("/")[-1]

                resource_id = f"{hosted_zone_id}:{vpc_id}:{vpc_region}"
                translated_resource = hub.tool.aws.route53.conversion_utils.convert_raw_zone_association_to_present(
                    hosted_zone_id=hosted_zone_id,
                    vpc_id=vpc_id,
                    vpc_region=vpc_region,
                    idem_resource_name=resource_id,
                )

                result[resource_id] = {
                    "aws.route53.hosted_zone_association.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in translated_resource.items()
                    ]
                }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS API Gateway v2 Authorizer resources.
"""


def convert_raw_authorizer_to_present(
    hub, api_id: str, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 Authorizer resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        api_id(string): The API resource identifier in Amazon Web Services.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "AuthorizerCredentialsArn": "authorizer_credentials_arn",
            "AuthorizerPayloadFormatVersion": "authorizer_payload_format_version",
            "AuthorizerResultTtlInSeconds": "authorizer_result_ttl_in_seconds",
            "AuthorizerType": "authorizer_type",
            "AuthorizerUri": "authorizer_uri",
            "EnableSimpleResponses": "enable_simple_responses",
            "IdentitySource": "identity_source",
            "JwtConfiguration": "jwt_configuration",
        }
    )
    resource_translated = {
        "name": raw_resource.get("Name"),
        "resource_id": raw_resource.get("AuthorizerId"),
        "api_id": api_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    api_id: str,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
) -> Dict[str, Any]:
    r"""
    Updates an AWS API Gateway v2 Authorizer resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Authorizer resource identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "authorizer_credentials_arn": "AuthorizerCredentialsArn",
            "authorizer_payload_format_version": "AuthorizerPayloadFormatVersion",
            "authorizer_result_ttl_in_seconds": "AuthorizerResultTtlInSeconds",
            "authorizer_type": "AuthorizerType",
            "authorizer_uri": "AuthorizerUri",
            "enable_simple_responses": "EnableSimpleResponses",
            "identity_source": "IdentitySource",
            "jwt_configuration": "JwtConfiguration",
        }
    )

    parameters_to_update = {}

    identity_source = resource_parameters.get("identity_source")
    if identity_source is not None:
        if set(identity_source) != set(raw_resource.get("identity_source", [])):
            parameters_to_update["IdentitySource"] = identity_source
        resource_parameters.pop("identity_source")

    jwt_configuration = resource_parameters.get("jwt_configuration")
    if jwt_configuration is not None:
        update_jwt_configuration = False
        old_jwt_configuration = raw_resource.get("jwt_configuration")

        if set(jwt_configuration.get("Audience", [])) != set(
            old_jwt_configuration.get("Audience", [])
        ):
            update_jwt_configuration = True
        elif jwt_configuration.get("Issuer") is not None and jwt_configuration.get(
            "Issuer"
        ) != old_jwt_configuration.get("Issuer"):
            update_jwt_configuration = True

        if update_jwt_configuration:
            parameters_to_update["JwtConfiguration"] = jwt_configuration

        resource_parameters.pop("jwt_configuration")

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource.get(key):
            parameters_to_update[parameters[key]] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_present, parameter_raw in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                f"Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_authorizer(
                ctx,
                ApiId=api_id,
                AuthorizerId=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                f"Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result



"""

hub.exec.boto3.client.elbv2.create_target_group
hub.exec.boto3.client.elbv2.delete_target_group
hub.exec.boto3.client.elbv2.deregister_targets
hub.exec.boto3.client.elbv2.describe_target_group_attributes
hub.exec.boto3.client.elbv2.describe_target_groups
hub.exec.boto3.client.elbv2.describe_target_health
hub.exec.boto3.client.elbv2.modify_target_group
hub.exec.boto3.client.elbv2.modify_target_group_attributes
hub.exec.boto3.client.elbv2.register_targets
hub.exec.boto3.client.elbv2.remove_tags
hub.exec.boto3.client.elbv2.describe_tags
hub.exec.boto3.client.elbv2.add_tags


"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    protocol: str = None,
    protocol_version: str = None,
    port: int = None,
    vpc_id: str = None,
    health_check_protocol: str = None,
    health_check_port: str = None,
    health_check_enabled: bool = None,
    health_check_path: str = None,
    health_check_interval_seconds: int = None,
    health_check_timeout_seconds: int = None,
    healthy_threshold_count: int = None,
    unhealthy_threshold_count: int = None,
    matcher: make_dataclass(
        "SubnetMapping",
        [
            ("HttpCode", str, field(default=None)),
            ("GrpcCode", str, field(default=None)),
        ],
    ) = None,
    target_type: str = None,
    tags: Dict[str, str] = None,
    ip_address_type: str = None,
    attributes: make_dataclass(
        "Attribute",
        [
            ("Key", str, field(default=None)),
            ("Value", str, field(default=None)),
        ],
    ) = None,
    targets: List[
        make_dataclass(
            "Target",
            [
                ("Id", str, field(default=None)),
                ("Port", int, field(default=None)),
                ("AvailabilityZone", str, field(default=None)),
            ],
        )
    ] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Creates a target group. This operation is idempotent, which means that it completes at most one time. If you
    attempt to create multiple target groups with the same settings, each call succeeds.

    Args:
        hub:
        ctx:
        name(Text): An Idem name of the resource.
            This name must be unique per region per account, can have a maximum of 32 characters, must contain only
            alphanumeric characters or hyphens, and must not begin or end with a hyphen.
        protocol(Text, optional): The protocol to use for routing traffic to the targets.
            1. For Application Load Balancers, the supported protocols are HTTP and HTTPS.
            2. For Network Load Balancers, the supported protocols are TCP, TLS, UDP, or TCP_UDP.
            3. For Gateway Load Balancers, the supported protocol is GENEVE.
            4. A TCP_UDP listener must be associated with a TCP_UDP target group. If the target is a Lambda function,
               this parameter does not apply.
        protocol_version(Text, optional):  [HTTP/HTTPS protocol] The protocol version. Specify GRPC to send requests to
            targets using gRPC. Specify HTTP2 to send requests to targets using HTTP/2. The default is HTTP1 , which
            sends requests to targets using HTTP/1.1.
        port(Integer, optional): The port on which the targets receive traffic. This port is used unless you specify a
            port override when registering the target. If the target is a Lambda function, this parameter does not
            apply. If the protocol is GENEVE, the supported port is 6081.
        vpc_id(Text, optional): The identifier of the virtual private cloud (VPC). If the target is a Lambda function,
            this parameter does not apply. Otherwise, this parameter is required.
        health_check_protocol(Text, optional): The protocol the load balancer uses when performing health checks on
            targets. For Application Load Balancers, the default is HTTP. For Network Load Balancers and Gateway
            Load Balancers, the default is TCP. The TCP protocol is not supported for health checks if the protocol of
            the target group is HTTP/HTTPS. GENEVE, TLS, UDP, and TCP_UDP protocols are not supported for health checks.
        health_check_port(Text, optional): The port the load balancer uses when performing health checks on targets.
            If the protocol is HTTP, HTTPS, TCP, TLS, UDP, or TCP_UDP, the default is traffic-port , which is the port
            on which each target receives traffic from the load balancer. If the protocol is GENEVE, default is port 80.
        health_check_enabled(Boolean, optional): Indicates whether health checks are enabled. If the target type is
            lambda , health checks are disabled by default but can be enabled. If the target type is instance , ip ,
            or alb , health checks are always enabled and cannot be disabled.
        health_check_path(Text, optional): [HTTP/HTTPS health checks] The destination for health checks on the targets.
            [HTTP1 or HTTP2 protocol version] The ping path. The default is /.
            [GRPC protocol version] The path of a custom health check method with the format /package.service/method.
            The default is /Amazon Web Services.ALB/healthcheck.
        health_check_interval_seconds(Integer, optional): The approximate amount of time, in seconds, between health
            checks of an individual target. If the target group protocol is HTTP or HTTPS, the default is 30 seconds.
            If the target group protocol is TCP, TLS, UDP, or TCP_UDP, the supported values are 10 and 30 seconds and
            the default is 30 seconds. If the target group protocol is GENEVE, the default is 10 seconds. If the target
            type is lambda , the default is 35 seconds.
        health_check_timeout_seconds(Integer, optional): The amount of time, in seconds, during which no response from a
            target means a failed health check. For target groups with a protocol of HTTP/HTTPS/GENEVE, the default is
            5 seconds. For target groups with a protocol of TCP or TLS, this value must be 6 seconds for HTTP health
            checks and 10 seconds for TCP & HTTPS health checks. If the target type is lambda, default is 30 seconds.
        healthy_threshold_count(Integer, optional): The number of consecutive health checks successes required before
            considering an unhealthy target healthy. For target groups with a protocol of HTTP/HTTPS, the default is 5.
            For target groups with a protocol of TCP/TLS/GENEVE, default is 3. If the target type is lambda, default = 5.
        unhealthy_threshold_count(Integer, optional): The number of consecutive health check failures required before
            considering a target unhealthy. If the target group protocol is HTTP or HTTPS, the default is 2. If the
            target group protocol is TCP or TLS, this value must be the same as the healthy threshold count. If the
            target group protocol is GENEVE, the default is 3. If the target type is lambda , the default is 2.
        matcher(Dict[str, str]), optional): [HTTP/HTTPS health checks] The HTTP or gRPC codes to use when checking for a
            successful response from a target.
            ** HttpCode(Text) - For Application Load Balancers, you can specify values between 200 and 499, and the
                default value is 200. You can specify multiple values (for example, "200,202") or a range of values
                (for example, "200-299").
                For Network Load Balancers and Gateway Load Balancers, this must be "200–399".
                Note that when using shorthand syntax, some values such as commas need to be escaped.

            ** GrpcCode(Text) - You can specify values between 0 and 99. You can specify multiple values (for example,
                "0,1") or a range of values (for example, "0-5"). The default value is 12.
        target_type(Text, optional): The type of target that you must specify when registering targets with this
            target group. You can't specify targets for a target group using more than one target type.
            ** instance - Register targets by instance ID. This is the default value.
            ** ip - Register targets by IP address. You can specify IP addresses from the subnets of the virtual private
                cloud (VPC) for the target group, the RFC 1918 range (10.0.0.0/8, 172.16.0.0/12, and 192.168.0.0/16),
                and the RFC 6598 range (100.64.0.0/10). You can't specify publicly routable IP addresses.
            ** lambda - Register a single Lambda function as a target.
            ** alb - Register a single Application Load Balancer as a target.
        tags([Dict[str, str]], optional): A list of tags to assign to the load balancer. For more information about
            tagging your load balancer, see Tag Your Classic Load Balancer in the Classic Load Balancers Guide.
            Defaults to None.
            * Key (Text): The key of the tag.
            * Value (Text, optional): The value of the tag.
        ip_address_type(Text, optional): The type of IP address used for this target group. The possible values are ipv4
            and ipv6 . This is an optional parameter. If not specified, the IP address type defaults to ipv4 .
        attributes(Dict[str, str], optional): The load balancer attributes.
            Key (Text): The name of the attribute.

            The following attribute is supported by all load balancers:
                1. 'deregistration_delay.timeout_seconds' - The amount of time, in seconds, for Elastic Load Balancing
                    to wait before changing the state of a deregistering target from draining to unused . The range is
                    0-3600 seconds. The default value is 300 seconds. If the target is a Lambda function, this
                    attribute is not supported.

            The following attributes are supported by both Application LoadBalancers and Network Load Balancers:
                1. 'stickiness.enabled' - Indicates whether sticky sessions are enabled.
                    The value is true or false. The default is false .
                2. 'stickiness.type' - The type of sticky sessions. The possible values are lb_cookie and app_cookie for
                    Application Load Balancers or source_ip for Network Load Balancers.

            The following attributes are supported only if the load balancer is an Application Load Balancer and the
                target is an instance or an IP address:
                1. 'load_balancing.algorithm.type' - The load balancing algorithm determines how the load balancer
                    selects targets when routing requests. The value is round_robin or least_outstanding_requests.
                    The default is round_robin .
                2. 'slow_start.duration_seconds' - The time period, in seconds, during which a newly registered target
                    receives an increasing share of the traffic to the target group. After this time period ends, the
                    target receives its full share of traffic. The range is 30-900 seconds (15 minutes).
                    The default is 0 seconds (disabled).
                3. 'stickiness.app_cookie.cookie_name' - Indicates the name of the application-based cookie. Names that
                    start with the following prefixes are not allowed: AWSALB , AWSALBAPP , and AWSALBTG ; they're
                    reserved for use by the load balancer.
                4. 'stickiness.app_cookie.duration_seconds' - The time period, in seconds, during which requests from a
                    client should be routed to the same target. After this time period expires, the application-based
                    cookie is considered stale. The range is 1 second to 1 week (604800 seconds).
                    The default value is 1 day (86400 seconds).
                5. 'stickiness.lb_cookie.duration_seconds' - The time period, in seconds, during which requests from a
                    client should be routed to the same target. After this time period expires, the load
                    balancer-generated cookie is considered stale. The range is 1 second to 1 week (604800 seconds).
                    The default value is 1 day (86400 seconds).

            The following attribute is supported only if the load balancer is an Application Load Balancer and the
                target is a Lambda function:
                1. 'lambda.multi_value_headers.enabled' - Indicates whether the request and response headers that are
                    exchanged between the load balancer and the Lambda function include arrays of values or strings.
                    The value is true or false. The default is false. If the value is false and the request contains
                    a duplicate header field name or query parameter key, the load balancer uses the last value sent by
                    the client.

            The following attributes are supported only by Network Load Balancers:
                1. 'deregistration_delay.connection_termination.enabled' - Indicates whether the load balancer
                    terminates connections at the end of the deregistration timeout. The value is true or false .
                    The default is false .
                2. 'preserve_client_ip.enabled' - Indicates whether client IP preservation is enabled. The value is
                    true or false . The default is disabled if the target group type is IP address and the target group
                    protocol is TCP or TLS. Otherwise, the default is enabled. Client IP preservation cannot be
                    disabled for UDP and TCP_UDP target groups.
                3. 'proxy_protocol_v2.enabled' - Indicates whether Proxy Protocol version 2 is enabled. The value is
                    true or false . The default is false .

        targets(List[Dict[str, Any]], optional): The targets. If you specified a port override when you registered a
            target, you must specify both the target ID and the port when you de-register it.

            Information about a target.
            Id(Text): The ID of the target. If the target type of the target group is instance, specify an instance
                ID. If the target type is ip , specify an IP address. If the target type is lambda , specify the ARN of
                the Lambda function. If the target type is alb , specify the ARN of the Application Load Balancer target.
            Port(Integer): The port on which the target is listening. If the target group protocol is GENEVE, the
                supported port is 6081. If the target type is alb , the targeted Application Load Balancer must have at
                least one listener whose port matches the target group port. Not used if the target is a Lambda function.
            AvailabilityZone(Text): An Availability Zone or all . This determines whether the target receives traffic
                from the load balancer nodes in the specified Availability Zone or from all enabled Availability Zones
                for the load balancer.
                1. This parameter is not supported if the target type of the target group is instance or alb .
                2. If the target type is ip and the IP address is in a subnet of the VPC for the target group, the
                   Availability Zone is automatically detected and this parameter is optional. If the IP address is
                   outside the VPC, this parameter is required.
                3. With an Application Load Balancer, if the target type is ip and the IP address is outside the VPC for
                   the target group, the only supported value is all .
                4. If the target type is lambda , this parameter is optional and the only supported value is all .
        resource_id:(Text, optional): The Amazon Resource Name (ARN) of the ElasticLoadBalancingv2 target group.

    Request Syntax:
        [target_group-name]:
          aws.elbv2.target_group.present:
            - name: 'string'
            - protocol: 'string',
            - protocol_version: 'string',
            - port: 'string',
            - vpc_id: 'string',
            - health_check_protocol: 'string',
            - health_check_port: 'string',
            - health_check_enabled: 'boolean',
            - health_check_path: 'string',
            - health_check_interval_seconds: 'integer',
            - health_check_timeout_seconds: 'integer',
            - healthy_threshold_count: 'integer',
            - unhealthy_threshold_count: 'integer',
            - matcher: list,
            - target_type: 'string',
            - tags: dict,
            - ip_address_type: 'string',
            - resource_id: 'string',
            - targets: 'list',

    Returns:
        Dict[str, Any]

    Examples:

    .. code-block:: sls

        test_target-group_name:
          aws.elbv2.target_group.present:
            - name: my-target-group
            - protocol: HTTP,
            - protocol_version: HTTP2,
            - port: 80,
            - vpc_id: vpc-3ac0fb5f,
            - health_check_protocol: HTTP,
            - health_check_port: traffic-port,
            - health_check_enabled: True,
            - health_check_path: /health-check,
            - health_check_interval_seconds: 30,
            - health_check_timeout_seconds: 60,
            - healthy_threshold_count: 5,
            - unhealthy_threshold_count: 2,
            - matcher:
              - HttpCode: 200
            - target_type: instance,
            - tags:
                name: target-group-name
            - ip_address_type: ipv4,
            - targets:
              - Id: i-80c8dd94
                Port: 80
                AvailabilityZone: us-east-1a
            - resource_id: arn:aws:elasticloadbalancing:us-west-2:123456789012:targetgroup/my-targets/73e2d6bc24d8a067

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.elbv2.target_group.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        input_map = {
            "protocol": protocol,
            "port": port,
            "vpc_id": vpc_id,
            "health_check_protocol": health_check_protocol,
            "health_check_port": health_check_port,
            "health_check_enabled": health_check_enabled,
            "health_check_interval_seconds": health_check_interval_seconds,
            "health_check_timeout_seconds": health_check_timeout_seconds,
            "healthy_threshold_count": healthy_threshold_count,
            "unhealthy_threshold_count": unhealthy_threshold_count,
            "health_check_path": health_check_path,
            "matcher": matcher,
            "target_type": target_type,
            "protocol_version": protocol_version,
            "ip_address_type": ip_address_type,
            "tags": tags,
            "attributes": attributes,
            "targets": targets,
        }
        update_return = await hub.tool.aws.elbv2.target_group.update(
            ctx=ctx,
            name=name,
            current_state=before["ret"],
            input_map=input_map,
            plan_state=plan_state,
            resource_id=resource_id,
        )
        result["comment"] = update_return["comment"]
        if not update_return["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_return["ret"])
        if resource_updated and ctx.get("test", False):
            result["comment"] += list(
                hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.elbv2.target_group", name=name
                )
            )

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags = await hub.tool.aws.elbv2.tag.update(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"]["tags"],
                new_tags=tags,
            )
            result["comment"] += list(update_tags["comment"])
            result["result"] = update_tags["result"]
            resource_updated = resource_updated or bool(update_tags["ret"])
            if ctx.get("test", False) and update_tags["ret"]:
                plan_state["tags"] = update_tags["ret"]

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "protocol": protocol,
                    "protocol_version": protocol_version,
                    "port": port,
                    "vpc_id": vpc_id,
                    "health_check_protocol": health_check_protocol,
                    "health_check_port": health_check_port,
                    "health_check_enabled": health_check_enabled,
                    "health_check_path": health_check_path,
                    "health_check_interval_seconds": health_check_interval_seconds,
                    "health_check_timeout_seconds": health_check_timeout_seconds,
                    "healthy_threshold_count": healthy_threshold_count,
                    "unhealthy_threshold_count": unhealthy_threshold_count,
                    "matcher": matcher,
                    "target_type": target_type,
                    "tags": tags,
                    "targets": targets,
                    "attributes": attributes,
                    "ip_address_type": ip_address_type,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elbv2.target_group", name=name
                )
            )
            return result
        ret = await hub.exec.boto3.client.elbv2.create_target_group(
            ctx,
            Name=name,
            Protocol=protocol,
            ProtocolVersion=protocol_version,
            Port=port,
            VpcId=vpc_id,
            HealthCheckProtocol=health_check_protocol,
            HealthCheckPort=health_check_port,
            HealthCheckEnabled=health_check_enabled,
            HealthCheckPath=health_check_path,
            HealthCheckIntervalSeconds=health_check_interval_seconds,
            HealthCheckTimeoutSeconds=health_check_timeout_seconds,
            HealthyThresholdCount=healthy_threshold_count,
            UnhealthyThresholdCount=unhealthy_threshold_count,
            Matcher=matcher,
            TargetType=target_type,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            IpAddressType=ip_address_type,
        )
        if not ret["result"]:
            result["comment"] = list(ret["comment"])
            result["result"] = False
            return result
        resource_id = (ret["ret"]["TargetGroups"][0]).get("TargetGroupArn")
        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elbv2.target_group", name=name
            )
        )
        if targets:
            register_targets = await hub.exec.boto3.client.elbv2.register_targets(
                ctx,
                TargetGroupArn=resource_id,
                Targets=targets,
            )
            if not register_targets["result"]:
                result["comment"] += list(register_targets["comment"])
                result["result"] = False
                return result
            result["comment"].append("Registered Targets.")
        if attributes:
            modify_attributes = (
                await hub.exec.boto3.client.elbv2.modify_target_group_attributes(
                    ctx,
                    TargetGroupArn=resource_id,
                    Attributes=attributes,
                )
            )
            if not modify_attributes["result"]:
                result["comment"] += list(modify_attributes["comment"])
                result["result"] = False
                return result
            result["comment"].append("Added Attributes.")

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or resource_updated:
        after = await hub.exec.aws.elbv2.target_group.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] += list(after["comment"])
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    1. Deletes the specified target group.
    2. You can delete a target group if it is not referenced by any actions. Deleting a target group also deletes any
       associated health checks. Deleting a target group does not affect its registered targets. For example, any EC2
       instances continue to run until you stop or terminate them.

    Args:
        name(Text): Idem name of the load balancer.
        resource_id(Text, optional): The Amazon Resource Name (ARN) of ElasticLoadBalancingv2 Target Group.

    Request Syntax:
        [target_group-name]:
          aws.elbv2.target_group.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

        Examples:

        .. code-block:: sls

            test_load-balancer_name:
                aws.elbv2.target_group.absent:
                  - name: my-load-balancer
                  - resource_id: arn:aws:elasticloadbalancing:us-west-2:123456789012:targetgroup/my-targets/73e2d6bc24d8a067

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elbv2.target_group", name=name
            )
        )
        return result
    before = await hub.exec.aws.elbv2.target_group.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elbv2.target_group", name=name
            )
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.elbv2.target_group", name=name
                )
            )
            return result
        else:
            ret = await hub.exec.boto3.client.elbv2.delete_target_group(
                ctx, TargetGroupArn=resource_id
            )
            if not ret["result"]:
                result["comment"] = list(before["comment"])
                result["result"] = False
                return result
            result["comment"] = list(
                hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.elbv2.target_group", name=name
                )
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describes the resource in a way that can be recreated/managed with the corresponding "present" function
    By default, all target groups are described.

    Returns:
        Dict[str, Any]

    Examples:

    .. code-block:: bash

        $ idem describe aws.elbv2.target_group
    """
    result = {}
    ret = await hub.exec.boto3.client.elbv2.describe_target_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.elbv2.target_group: {ret['comment']}")
        return result
    if ret["ret"].get("TargetGroups"):
        for target_group in ret["ret"].get("TargetGroups"):
            tags_ret = await hub.exec.boto3.client.elbv2.describe_tags(
                ctx, ResourceArns=[target_group.get("TargetGroupArn")]
            )
            if not tags_ret["result"]:
                # Error fetching tags. If fetching tags itself fails, just skip this and continue. Since tags are
                # optional, they need not be associated with all load balancers. But API needs to succeed.
                hub.log.warning(
                    f"Failed listing tags for aws.elbv2.target_group '{target_group.get('TargetGroupArn')}'"
                    f"Describe will skip this aws.elbv2.target_group and continue."
                )
                continue

            tags = []
            if tags_ret.get("ret") and tags_ret.get("ret")["TagDescriptions"]:
                tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    tags_ret.get("ret").get("TagDescriptions")[0].get("Tags")
                )

            attributes_ret = (
                await hub.exec.boto3.client.elbv2.describe_target_group_attributes(
                    ctx, TargetGroupArn=target_group.get("TargetGroupArn")
                )
            )
            if not attributes_ret["result"]:
                # Error fetching attributes (Attributes are optional. Not every load_balancer needs to have them.)
                hub.log.warning(
                    f"Failed listing attributes for aws.elbv2.target_group '{target_group.get('TargetGroupArn')}'"
                    f"Describe will skip this aws.elbv2.target_group and continue. "
                )
                continue

            attributes = []
            if attributes_ret.get("ret") and attributes_ret.get("ret").get(
                "Attributes"
            ):
                attributes = attributes_ret["ret"].get("Attributes")

            targets_ret = await hub.exec.boto3.client.elbv2.describe_target_health(
                ctx, TargetGroupArn=target_group.get("TargetGroupArn")
            )
            if not targets_ret["result"]:
                # Error fetching attributes (Attributes are optional. Not every load_balancer needs to have them.)
                hub.log.warning(
                    f"Failed listing targets for aws.elbv2.target_group '{target_group.get('TargetGroupArn')}'"
                    f"Describe will skip this aws.elbv2.target_group and continue. "
                )
                continue

            targets = []
            if targets_ret.get("ret") and targets_ret.get("ret").get(
                "TargetHealthDescriptions"
            ):
                targets = targets_ret.get("ret").get("TargetHealthDescriptions")

            resource_converted = (
                hub.tool.aws.elbv2.conversion_utils.convert_raw_target_group_to_present(
                    raw_resource=target_group,
                    idem_resource_name=target_group.get("TargetGroupName"),
                    tags=tags,
                    attributes=attributes,
                    targets=targets,
                )
            )
            result[resource_converted["resource_id"]] = {
                "aws.elbv2.target_group.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted.items()
                ]
            }
    return result



import copy
from typing import Any
from typing import Dict


async def get(hub, ctx, name, resource_id: str = None):
    r"""
    Provides details about a specific instance profile as a data-source.

    Args:
        name(Text): The AWS Instance Profile name.
        resource_id(Text, Optional): AWS Instance Profile name to identify the resource.
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.iam.get_instance_profile(
        ctx, InstanceProfileName=resource_id if resource_id else name
    )

    if not ret["result"]:
        if "NoSuchEntity" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.instance_profile", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    resource = ret["ret"]["InstanceProfile"]
    result[
        "ret"
    ] = hub.tool.aws.iam.conversion_utils.convert_raw_instance_profile_to_present(
        resource
    )
    return result


async def update_instance_profile_tags(
    hub,
    ctx,
    instance_profile_name: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """
    Update tags of AWS IAM Instance Profile

    TODO - this method might fail with localstack but is successful with a real AWS account

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        instance_profile_name: AWS IAM instance profile name
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": "dict tags after update"}
    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_instance_profile(
                ctx,
                InstanceProfileName=instance_profile_name,
                TagKeys=list(tags_to_remove.keys()),
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_instance_profile(
                ctx,
                InstanceProfileName=instance_profile_name,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



"""Exec module for managing Amazon Elasticache Replication Groups."""
from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict:
    """Retrieves the specified AWS Elasticache Replication Group.

    Arg:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            The ID of the replication group in Amazon Web Services.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The Elasticache Replication Group in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.elasticache.replication_group.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.elasticache.replication_group.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.elasticache.describe_replication_groups(
        ctx=ctx,
        ReplicationGroupId=resource_id,
    )
    if not ret["result"]:
        if "ReplicationGroupNotFoundFault" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.elasticache.replication_group", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] = list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["ReplicationGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
        )
        return result
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    resource = ret["ret"]["ReplicationGroups"][0]
    before_tag = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
        ctx, ResourceName=resource.get("ARN")
    )
    if not before_tag["result"]:
        result["result"] = False
        result["comment"] += list(before_tag["comment"])
        return result
    resource["Tags"] = before_tag["ret"].get("TagList", [])
    result[
        "ret"
    ] = hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_replication_group_to_present(
        ctx, raw_resource=resource, idem_resource_name=name
    )
    return result



from typing import Any
from typing import Dict


async def update_status(hub, ctx, tags: Dict[str, Any]):
    """
    Updates status of a cost allocation tag. If the tag status is already in same state, the request doesn't fail.
    Instead, it doesn't have any effect on the tag status (for example, activating the active tag).

    Args:
        tags(Dict): Dictionary of Cost allocation tags with status.

    Returns:
        Dict[str, str]

    """
    result = dict(comment=(), result=True, ret=None)

    # converting tag dict to list of tags, format required for api
    cost_allocation_tags = [
        {"TagKey": key, "Status": value} for key, value in tags.items()
    ]
    if not ctx.get("test", False):
        update_ret = await hub.exec.boto3.client.ce.update_cost_allocation_tags_status(
            ctx=ctx, CostAllocationTagsStatus=cost_allocation_tags
        )
        if not update_ret["result"]:
            result["comment"] += update_ret["comment"]
            result["result"] = False
            return result
    result["ret"] = {}
    result["ret"]["cost_allocation_tags"] = cost_allocation_tags
    if ctx.get("test", False):
        result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
            resource_type="aws.costexplorer.cost_allocation_tag",
            name=cost_allocation_tags,
        )
    else:
        result["comment"] += hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.costexplorer.cost_allocation_tag",
            name=cost_allocation_tags,
        )

    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict

from dict_tools import data


async def update_password_policy(
    hub,
    ctx,
    name: str,
    before: Dict[str, Any],
    policy_parameters: Dict[str, Any],
):
    """
    Updates the password policy settings for the Amazon Web Services account.

    Args:
        hub: required for functions in hub.
        ctx: context.
        name: Password policy name.
        before(Dict): Existing resource parameters in Amazon Web Services.
        policy_parameters(Dict): Password policy attributes.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    parameters_to_update = {}

    parameters = OrderedDict(
        {
            "minimum_password_length": "MinimumPasswordLength",
            "require_symbols": "RequireSymbols",
            "require_numbers": "RequireNumbers",
            "require_uppercase_characters": "RequireUppercaseCharacters",
            "require_lowercase_characters": "RequireLowercaseCharacters",
            "allow_users_to_change_password": "AllowUsersToChangePassword",
            "max_password_age": "MaxPasswordAge",
            "password_reuse_prevention": "PasswordReusePrevention",
            "hard_expiry": "HardExpiry",
        }
    )
    for key, value in parameters.items():
        if key in policy_parameters.keys():
            parameters_to_update[value] = policy_parameters[key]

    # expire_passwords can not be updated as direct param but is available in get
    old_policy = copy.deepcopy(before)
    old_policy.pop("expire_passwords")

    if not policy_parameters["max_password_age"]:
        policy_parameters.pop("max_password_age")
    if not policy_parameters["password_reuse_prevention"]:
        policy_parameters.pop("password_reuse_prevention")
    policy_diff = data.recursive_diff(policy_parameters, old_policy, ignore_order=True)
    if policy_diff:
        result["ret"] = {}

        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.iam.update_account_password_policy(
                ctx=ctx, **parameters_to_update
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.iam.password_policy", name=name
            )

        for key, value in parameters.items():
            if value in parameters_to_update:
                result["ret"][key] = parameters_to_update[value]
    return result



from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict:
    """
    Pass required params to get a load balancer resource.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS ELBv2 Load Balancer ARN to identify the resource.

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.elbv2.load_balancer.search_raw(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not ret["result"]:
        if "LoadBalancerNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.elbv2.load_balancer", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["LoadBalancers"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.elbv2.load_balancer", name=name
            )
        )
        return result

    resource_id = ret["ret"]["LoadBalancers"][0].get("LoadBalancerArn")
    tags = attributes = []
    if resource_id:
        tags_ret = await hub.exec.boto3.client.elbv2.describe_tags(
            ctx, ResourceArns=[resource_id]
        )

        if not tags_ret["result"]:
            result["comment"] = list(tags_ret["comment"])
        else:
            if tags_ret["result"] and tags_ret.get("ret")["TagDescriptions"]:
                tags = (tags_ret.get("ret").get("TagDescriptions")[0]).get("Tags")

        attributes_ret = (
            await hub.exec.boto3.client.elbv2.describe_load_balancer_attributes(
                ctx, LoadBalancerArn=resource_id
            )
        )
        if not attributes_ret.get("result"):
            result["comment"] += list(attributes_ret["comment"])
        else:
            if attributes_ret.get("ret") and attributes_ret.get("ret").get(
                "Attributes"
            ):
                attributes = attributes_ret["ret"].get("Attributes")
    result[
        "ret"
    ] = hub.tool.aws.elbv2.conversion_utils.convert_raw_load_balancer_to_present(
        raw_resource=ret["ret"]["LoadBalancers"][0],
        idem_resource_name=name,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
        attributes=attributes,
    )
    return result



from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str, scope: str) -> Dict[str, Any]:
    """
    Function for getting the wafv2 regular expression pattern set with name, Id and scope
    Returns the state of a wafv2 regular expression pattern set

    Args:
        name(str): The name of the regex pattern set.
        resource_id(str): An identifier of the resource in the provider. Defaults to None.
        scope(str): Specifies whether this is for an Amazon CloudFront distribution or for a regional application. A
            regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API,
            or an AppSync GraphQL API.  To work with CloudFront, you must also specify the Region US East
            (N. Virginia) as follows:    CLI - Specify the Region when you use the CloudFront scope:
            --scope=CLOUDFRONT --region=us-east-1.    API and SDKs - For all calls, use the Region endpoint
            us-east-1.

    Returns
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The WAFV2 regular expression pattern set in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.wafv2.regex_pattern_set.get name="idem_name" resource_id="resource_id" scope="REGIONAL"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.wafv2.regex_pattern_set.get(
                    ctx, name=name, resource_id=resource_id, scope="REGIONAL"
                )
    """
    result = dict(ret=None, comment=[], result=True)
    is_invalid_scope_and_region = (
        hub.tool.aws.wafv2.regex_pattern_set.check_if_invalid_scope_and_region(
            ctx["acct"].get("region_name"), scope
        )
    )
    if is_invalid_scope_and_region:
        result["result"] = False
        result["comment"].append(is_invalid_scope_and_region)
        return result
    ret = await hub.exec.boto3.client.wafv2.get_regex_pattern_set(
        ctx=ctx, Name=name, Scope=scope, Id=resource_id
    )
    if not ret["result"]:
        if "WAFNonexistentItemException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.wafv2.regex_pattern_set", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["RegexPatternSet"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.wafv2.regex_pattern_set", name=name
            )
        )
        return result

    resource = ret["ret"]["RegexPatternSet"]
    tags = []
    ret_tag = await hub.exec.boto3.client.wafv2.list_tags_for_resource(
        ctx, ResourceARN=resource["ARN"]
    )
    if not ret_tag["result"]:
        result["result"] = False
        result["comment"] = list(ret_tag["comment"])
        return result
    if (
        ret_tag["ret"]
        and ret_tag["ret"]["TagInfoForResource"]
        and ret_tag["ret"]["TagInfoForResource"].get("TagList")
    ):
        tags = ret_tag["ret"]["TagInfoForResource"].get("TagList")

    result[
        "ret"
    ] = hub.tool.aws.wafv2.conversion_utils.convert_raw_regex_pattern_set_to_present(
        ctx,
        raw_resource=resource,
        scope=scope,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
    )
    result["ret"]["LockToken"] = ret["ret"]["LockToken"]
    return result


async def list_(hub, ctx, name: str, scope: str) -> Dict:
    """
    Fetch a list of WAFV2 regex pattern sets
    Args:
        name(str): The name of the Idem state.
        scope(str): Specifies whether this is for an Amazon CloudFront distribution or for a regional application. A
            regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API,
            or an AppSync GraphQL API.  To work with CloudFront, you must also specify the Region US East
            (N. Virginia) as follows:    CLI - Specify the Region when you use the CloudFront scope:
            --scope=CLOUDFRONT --region=us-east-1.    API and SDKs - For all calls, use the Region endpoint
            us-east-1.

    Returns
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(list or None):
           List of WAFV2 regular expression pattern sets in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.wafv2.regex_pattern_set.list scope="REGIONAL"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.wafv2.regex_pattern_set.list(
                    ctx, scope="REGIONAL"
                )

    """
    result = dict(ret=[], comment=[], result=True)
    is_invalid_scope_and_region = (
        hub.tool.aws.wafv2.regex_pattern_set.check_if_invalid_scope_and_region(
            ctx["acct"].get("region_name"), scope
        )
    )
    if is_invalid_scope_and_region:
        result["result"] = False
        result["comment"].append(is_invalid_scope_and_region)
        return result

    ret = await hub.exec.boto3.client.wafv2.list_regex_pattern_sets(
        ctx=ctx, Scope=scope
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["RegexPatternSets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.wafv2.regex_pattern_set", name=name
            )
        )
        return result
    for regex_pattern_set in ret["ret"]["RegexPatternSets"]:
        get_regex_ret = await hub.exec.aws.wafv2.regex_pattern_set.get(
            ctx=ctx,
            name=regex_pattern_set["Name"],
            resource_id=regex_pattern_set["Id"],
            scope=scope,
        )
        if not get_regex_ret["result"]:
            result["comment"] += list(get_regex_ret["comment"])
            result["result"] = False
            return result
        result["ret"].append(get_regex_ret["ret"])
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.attach_user_policy
hub.exec.boto3.client.iam.detach_user_policy
hub.exec.boto3.client.iam.list_user_policies
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub, ctx, name: str, user_name: str, policy_arn: str, resource_id: str = None
) -> Dict[str, Any]:
    """Attaches the specified managed policy to the specified user. Use this operation to attach a managed policy to a user.

    Args:
        name(str):
            An Idem name of the state.

        user_name(str):
            The name (friendly name, not ARN) of the IAM user to detach the policy from. This parameter allows
            (through its regex pattern) a string of characters consisting of upper and lowercase alphanumeric
            characters with no spaces. You can also include any of the following characters: _+=,.@-

        policy_arn(str):
            The Amazon Resource Name (ARN) of the IAM policy you want to attach.

        resource_id(str, optional):
            An identifier refers to an existing resource. The format is <user_name>/<policy_arn>

    Request Syntax:
        .. code-block:: sls

            [iam-attach-user-policy]:
              aws.iam.user_policy_attachment.present:
                - name: "string"
                - user_name: 'string'
                - policy_arn: 'string'
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem-test-attach-policy:
              aws.iam.user_policy_attachment.present:
                - name: test-policy-attachment
                - user_name: serverless
                - policy_arn: arn:aws:iam::aws:policy/AdministratorAccess

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    generated_resource_id = f"{user_name}/{policy_arn}"
    if resource_id and generated_resource_id != resource_id:
        result["comment"].append(
            f"resource_id mismatches. Use newly generated resource_id"
        )
        # newly generated resource_id is used to populate resource state in get()
    before_ret = await hub.exec.aws.iam.user_policy_attachment.get(
        ctx, name=name, user_name=user_name, policy_arn=policy_arn
    )
    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result
    if before_ret["ret"]:
        result["comment"] += list(
            hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.iam.user_policy_attachment", name=name
            )
        )
        result["old_state"] = before_ret["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result
    else:
        if ctx.get("test", False):
            result["comment"] += list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.iam.user_policy_attachment", name=name
                )
            )
        else:
            ret = await hub.exec.boto3.client.iam.attach_user_policy(
                ctx,
                UserName=user_name,
                PolicyArn=policy_arn,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] += list(
                hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.iam.user_policy_attachment", name=name
                )
            )
        result["new_state"] = {
            "name": name,
            "user_name": user_name,
            "policy_arn": policy_arn,
            "resource_id": generated_resource_id,
        }
    return result


async def absent(
    hub,
    ctx,
    name: str,
    *,
    resource_id: str = None,
    user_name: str = None,
    policy_arn: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Removes the specified managed policy from the specified user.
    A user can also have inline policies embedded with it. To delete an inline policy, use DeleteUserPolicy

    Args:
        name(str):
            An Idem name of the state.

        resource_id(str, optional):
            An identifier refers to an existing resource. The format is <user_name>/<policy_arn> Either resource_id
            or both user_name and policy_arn should be specified for absent.

        user_name(str, optional):
            The name (friendly name, not ARN) of the IAM user to detach the policy from. This parameter allows
            (through its regex pattern ) a string of characters consisting of upper and lowercase alphanumeric
            characters with no spaces. You can also include any of the following characters: _+=,.@-

        policy_arn(str, optional):
            The Amazon Resource Name (ARN) of the IAM policy you want to attach.

    Request Syntax:
        .. code-block:: sls

            [iam-user-policy-name]:
              aws.iam.user_policy_attachment.absent:
                - name: 'string'
                - user_name: 'string'
                - policy_arn: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-attach-policy:
              aws.iam.user_policy_attachment.absent:
                - name: test-policy-attachment
                - user_name: serverless
                - policy_arn: arn:aws:iam::aws:policy/AdministratorAccess
    """

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    generated_resource_id = f"{user_name}/{policy_arn}"
    if resource_id and generated_resource_id != resource_id:
        result["comment"].append(
            f"resource_id mismatches. Use newly generated resource_id"
        )
    before_ret = await hub.exec.aws.iam.user_policy_attachment.get(
        ctx, name=name, user_name=user_name, policy_arn=policy_arn
    )
    if not before_ret["result"]:
        result["comment"] = before_ret["comment"]
        result["result"] = False
        return result
    if not before_ret["ret"]:
        result["comment"] += list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.iam.user_policy_attachment", name=name
            )
        )
    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] += list(
            hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.iam.user_policy_attachment", name=name
            )
        )
    else:
        result["old_state"] = before_ret["ret"]
        ret = await hub.exec.boto3.client.iam.detach_user_policy(
            ctx, UserName=user_name, PolicyArn=policy_arn
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] += list(ret["comment"])
            result["result"] = False
            return result
        result["comment"] += list(
            hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.iam.user_policy_attachment", name=name
            )
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists all managed policies that are attached to the specified IAM user. Lists all managed policies that are attached to the specified IAM user.



    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.iam.user_policy_attachment
    """

    result = {}
    # Fetch all users and then policies for each of them
    users = await hub.exec.boto3.client.iam.list_users(ctx)
    if not users["result"]:
        hub.log.debug(f"Could not describe user {users['comment']}")
        return {}
    user_names = [user.get("UserName") for user in users["ret"]["Users"]]
    for user_name in user_names:
        ret_attached_policies = (
            await hub.exec.boto3.client.iam.list_attached_user_policies(
                ctx=ctx, UserName=user_name
            )
        )
        if not ret_attached_policies["result"]:
            hub.log.warning(
                f"Could not get attached policy list with user {user_name} with error"
                f" {ret_attached_policies['comment']} . Describe will skip this user and continue."
            )
        else:
            resources = ret_attached_policies["ret"].get("AttachedPolicies")
            for resource in resources:
                resource_name = f"{user_name}-{resource.get('PolicyArn')}"
                resource_translated = []
                resource_translated.append({"name": user_name})
                resource_translated.append({"user_name": user_name})
                resource_translated.append({"policy_arn": resource.get("PolicyArn")})
                result[resource_name] = {
                    "aws.iam.user_policy_attachment.present": resource_translated
                }
    return result
